# -*- coding: utf-8 -*-
"""
Module for handling ObsPy RtTrace objects.

:copyright:
    The ObsPy Development Team (devs@obspy.org) & Anthony Lomax
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import copy
import warnings

import numpy as np

from obspy import Trace
from obspy.core import Stats
from obspy.realtime import signal
from obspy.realtime.rtmemory import RtMemory


# dictionary to map given type-strings to processing functions keys must be all
# lower case - values are tuples: (function name, number of RtMemory objects)
REALTIME_PROCESS_FUNCTIONS = {
    'scale': (signal.scale, 0),
    'offset': (signal.offset, 0),
    'integrate': (signal.integrate, 1),
    'differentiate': (signal.differentiate, 1),
    'boxcar': (signal.boxcar, 1),
    'tauc': (signal.tauc, 2),
    'mwpintegral': (signal.mwpintegral, 1),
    'kurtosis': (signal.kurtosis, 3),
}


class RtTrace(Trace):
    """
    An object containing data of a continuous series constructed dynamically
    from sequential data packets.

    New data packets may be periodically appended. Registered time-domain
    processes can be applied to the new data and the resulting trace will be
    left trimmed to maintain a specified maximum trace length.

    :type max_length: int, optional
    :param max_length: maximum trace length in seconds

    .. rubric:: Example

    RtTrace has been built to handle real time processing of periodically
    append data packets, such as adding and processing data requested from an
    SeedLink server. See :mod:`obspy.clients.seedlink` for further information.

    For the sake of simplicity we will just split data of an existing example
    file into multiple chucks (Trace objects) of about equal size (step 1 + 2)
    and append those chunks in a simple loop (step 4) into an RtTrace object.
    Additionally there are two real time processing functions registered to the
    RtTrace object (step 3) which will automatically process any appended data
    chunks.

    1. Read first trace of example SAC data file and extract contained time
       offset and epicentral distance of an earthquake::

        >>> from obspy.realtime import RtTrace
        >>> from obspy import read
        >>> from obspy.realtime.signal import calculate_mwp_mag
        >>> data_trace = read(
        ...     '/path/to/II.TLY.BHZ.SAC', round_sampling_interval=False)[0]
        >>> len(data_trace)
        12684
        >>> ref_time_offset = data_trace.stats.sac.a
        >>> print(ref_time_offset)
        301.506
        >>> epicentral_distance = data_trace.stats.sac.gcarc
        >>> print(epicentral_distance)
        30.0855

    2. Split given trace into a list of three sub-traces::

        >>> traces = data_trace / 3
        >>> [len(tr) for tr in traces]
        [4228, 4228, 4228]

    3. Assemble real time trace and register two processes::

        >>> rt_trace = RtTrace()
        >>> rt_trace.register_rt_process('integrate')
        1
        >>> rt_trace.register_rt_process('mwpintegral', mem_time=240,
        ...     ref_time=(data_trace.stats.starttime + ref_time_offset),
        ...     max_time=120, gain=1.610210e+09)
        2

    4. Append and auto-process packet data into RtTrace::

        >>> for tr in traces:
        ...     processed_trace = rt_trace.append(tr, gap_overlap_check=True)
        ...
        >>> len(rt_trace)
        12684

    5. Some post processing to get Mwp::

        >>> peak = np.amax(np.abs(rt_trace.data))
        >>> print(peak)
        0.136404
        >>> mwp = calculate_mwp_mag(peak, epicentral_distance)
        >>> print(mwp)  # doctest: +ELLIPSIS
        8.78902911791...
    """
    have_appended_data = False

    @classmethod
    def rt_process_functions_to_string(cls):
        """
        Return doc string for all predefined real-time processing functions.

        :rtype: str
        :return: String containing doc for all real-time processing functions.
        """
        string = 'Real-time processing functions (use as: ' + \
            'RtTrace.register_rt_process(process_name, [parameter values])):\n'
        for key in REALTIME_PROCESS_FUNCTIONS:
            string += '\n'
            string += '  ' + (str(key) + ' ' + 80 * '-')[:80]
            string += str(REALTIME_PROCESS_FUNCTIONS[key][0].__doc__)
        return string

    def __init__(self, max_length=None, *args, **kwargs):  # @UnusedVariable
        """
        Initializes an RtTrace.

        See :class:`obspy.core.trace.Trace` for all parameters.
        """
        # set window length attribute
        if max_length is not None and max_length <= 0:
            raise ValueError("Input max_length out of bounds: %s" % max_length)
        self.max_length = max_length

        # initialize processing list
        self.processing = []

        # initialize parent Trace with no data or header - all data must be
        # added using append
        super(RtTrace, self).__init__(data=np.array([]), header=None)

    def __eq__(self, other):
        """
        Implements rich comparison of RtTrace objects for "==" operator.

        Traces are the same, if both their data and stats are the same.
        """
        # check if other object is a RtTrace
        if not isinstance(other, RtTrace):
            return False
        return super(RtTrace, self).__eq__(other)

    def __add__(self, **kwargs):  # @UnusedVariable
        """
        Too ambiguous, throw an Error.

        .. seealso:: :meth:`obspy.realtime.rttrace.RtTrace.append`.
        """
        msg = "Too ambiguous for realtime trace data. Try: RtTrace.append()"
        raise NotImplementedError(msg)

    def append(self, trace, gap_overlap_check=False, verbose=False):
        """
        Appends a Trace object to this RtTrace.

        Registered real-time processing will be applied to copy of appended
        Trace object before it is appended.  This RtTrace will be truncated
        from the beginning to RtTrace.max_length, if specified.
        Sampling rate, data type and trace.id of both traces must match.

        :type trace: :class:`~obspy.core.trace.Trace`
        :param trace:  :class:`~obspy.core.trace.Trace` object to append to
            this RtTrace
        :type gap_overlap_check: bool, optional
        :param gap_overlap_check: Action to take when there is a gap or overlap
            between the end of this RtTrace and start of appended Trace:

            * If True, raise TypeError.
            * If False, all trace processing memory will be re-initialized to
              prevent false signal in processed trace.

            (default is ``True``).
        :type verbose: bool, optional
        :param verbose: Print additional information to stdout
        :return: NumPy :class:`~numpy.ndarray` object containing processed
            trace data from appended Trace object.
        """
        if not isinstance(trace, Trace):
            # only add Trace objects
            raise TypeError("Only obspy.core.trace.Trace objects are allowed")

        # sanity checks
        if self.have_appended_data:
            #  check id
            if self.get_id() != trace.get_id():
                raise TypeError("Trace ID differs:", self.get_id(),
                                trace.get_id())
            #  check sample rate
            if self.stats.sampling_rate != trace.stats.sampling_rate:
                raise TypeError("Sampling rate differs:",
                                self.stats.sampling_rate,
                                trace.stats.sampling_rate)
            #  check calibration factor
            if self.stats.calib != trace.stats.calib:
                raise TypeError("Calibration factor differs:",
                                self.stats.calib, trace.stats.calib)
            # check data type
            if self.data.dtype != trace.data.dtype:
                raise TypeError("Data type differs:",
                                self.data.dtype, trace.data.dtype)
        # TODO: IMPORTANT? Should improve check for gaps and overlaps
        # and handle more elegantly
        # check times
        gap_or_overlap = False
        if self.have_appended_data:
            # delta = int(math.floor(\
            #    round((rt.stats.starttime - lt.stats.endtime) * sr, 5) )) - 1
            diff = trace.stats.starttime - self.stats.endtime
            delta = diff * self.stats.sampling_rate - 1.0
            if verbose:
                msg = "%s: Overlap/gap of (%g) samples in data: (%s) (%s) " + \
                    "diff=%gs  dt=%gs"
                print(msg % (self.__class__.__name__,
                             delta, self.stats.endtime, trace.stats.starttime,
                             diff, self.stats.delta))
            if delta < -0.1:
                msg = "Overlap of (%g) samples in data: (%s) (%s) diff=%gs" + \
                    "  dt=%gs"
                msg = msg % (-delta, self.stats.endtime, trace.stats.starttime,
                             diff, self.stats.delta)
                if gap_overlap_check:
                    raise TypeError(msg)
                gap_or_overlap = True
            if delta > 0.1:
                msg = "Gap of (%g) samples in data: (%s) (%s) diff=%gs" + \
                    "  dt=%gs"
                msg = msg % (delta, self.stats.endtime, trace.stats.starttime,
                             diff, self.stats.delta)
                if gap_overlap_check:
                    raise TypeError(msg)
                gap_or_overlap = True
            if gap_or_overlap:
                msg += " - Trace processing memory will be re-initialized."
                warnings.warn(msg, UserWarning)
            else:
                # correct start time to pin absolute trace timing to start of
                # appended trace, this prevents slow drift of nominal trace
                # timing from absolute time when nominal sample rate differs
                # from true sample rate
                self.stats.starttime = \
                    self.stats.starttime + diff - self.stats.delta
                if verbose:
                    print("%s: self.stats.starttime adjusted by: %gs"
                          % (self.__class__.__name__, diff -
                             self.stats.delta))
        # first apply all registered processing to Trace
        for proc in self.processing:
            process_name, options, rtmemory_list = proc
            # if gap or overlap, clear memory
            if gap_or_overlap and rtmemory_list is not None:
                for n in range(len(rtmemory_list)):
                    rtmemory_list[n] = RtMemory()
            # apply processing
            trace = trace.copy()
            dtype = trace.data.dtype
            if hasattr(process_name, '__call__'):
                # check if direct function call
                trace.data = process_name(trace.data, **options)
            else:
                # got predefined function
                func = REALTIME_PROCESS_FUNCTIONS[process_name.lower()][0]
                options['rtmemory_list'] = rtmemory_list
                trace.data = func(trace, **options)
            # assure dtype is not changed
            trace.data = np.require(trace.data, dtype=dtype)
        # if first data, set stats
        if not self.have_appended_data:
            self.data = np.array(trace.data)
            self.stats = Stats(header=trace.stats)
            self.have_appended_data = True
            return trace
        # handle all following data sets
        # fix Trace.__add__ parameters
        # TODO: IMPORTANT? Should check for gaps and overlaps and handle
        # more elegantly
        sum_trace = Trace.__add__(
            self, trace, method=0, interpolation_samples=0,
            fill_value='latest', sanity_checks=True)
        # Trace.__add__ returns new Trace, so update to this RtTrace
        self.data = sum_trace.data
        # left trim if data length exceeds max_length
        if self.max_length is not None:
            max_samples = int(self.max_length * self.stats.sampling_rate + 0.5)
            if np.size(self.data) > max_samples:
                starttime = self.stats.starttime + \
                    (np.size(self.data) - max_samples) / \
                    self.stats.sampling_rate
                self._ltrim(starttime, pad=False, nearest_sample=True,
                            fill_value=None)
        return trace

    def register_rt_process(self, process, **options):
        """
        Adds real-time processing algorithm to processing list of this RtTrace.

        Processing function must be one of:
            %s. % REALTIME_PROCESS_FUNCTIONS.keys()
            or a non-recursive, time-domain NumPy or ObsPy function which takes
            a single array as an argument and returns an array

        :type process: str or callable
        :param process: Specifies which processing function is added,
            e.g. ``"boxcar"`` or ``np.abs``` (functions without brackets).
            See :mod:`obspy.realtime.signal` for all predefined processing
            functions.
        :type options: dict, optional
        :param options: Required keyword arguments to be passed the respective
            processing function, e.g. ``width=100`` for ``'boxcar'`` process.
            See :mod:`obspy.realtime.signal` for all options.
        :rtype: int
        :return: Length of processing list after registering new processing
            function.
        """
        # create process_name either from string or function name
        process_name = ("%s" % process).lower()

        # set processing entry for this process
        entry = False
        rtmemory_list = None
        if hasattr(process, '__call__'):
            # direct function call
            entry = (process, options, None)
        elif process_name in REALTIME_PROCESS_FUNCTIONS:
            # predefined function
            num = REALTIME_PROCESS_FUNCTIONS[process_name][1]
            if num:
                # make sure we have num new RtMemory instances
                rtmemory_list = [RtMemory() for _i in range(num)]
            entry = (process_name, options, rtmemory_list)
        else:
            # check if process name is contained within a predefined function,
            # e.g. 'int' for 'integrate'
            for key in REALTIME_PROCESS_FUNCTIONS:
                if not key.startswith(process_name):
                    continue
                process_name = key
                num = REALTIME_PROCESS_FUNCTIONS[process_name][1]
                if num:
                    # make sure we have num new RtMemory instances
                    rtmemory_list = [RtMemory() for _i in range(num)]
                entry = (process_name, options, rtmemory_list)
                break

        if not entry:
            raise NotImplementedError("Can't register process %s" % (process))

        # add process entry
        self.processing.append(entry)

        # add processing information to the stats dictionary
        proc_info = "realtime_process:%s:%s" % (process_name, options)
        self._internal_add_processing_info(proc_info)

        return len(self.processing)

    def copy(self, *args, **kwargs):
        """
        Returns a deepcopy of this RtTrace.
        """
        # XXX: ugly hack to allow deepcopy of an RtTrace object containing
        # registered NumPy function (numpy.ufunc) calls
        temp = copy.copy(self.processing)
        self.processing = []
        new = copy.deepcopy(self, *args, **kwargs)
        new.processing = temp
        return new


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
Base utilities and constants for ObsPy.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import functools
import glob
import importlib
import importlib.metadata
import inspect
import io
import os
from contextlib import contextmanager
from io import IOBase, TextIOBase, TextIOWrapper
from pathlib import Path
import re
import sys
import tempfile
import unicodedata
import warnings
from collections import OrderedDict
from pathlib import PurePath

import numpy as np

from obspy.core.util.attribdict import (AttribDict)
from obspy.core.util.deprecation_helpers import ObsPyDeprecationWarning
from obspy.core.util.misc import to_int_or_zero, buffered_load_entry_point


# defining ObsPy modules currently used by runtests and the path function
DEFAULT_MODULES = ['clients.filesystem', 'core', 'geodetics', 'imaging',
                   'io.ah', 'io.alsep', 'io.arclink', 'io.ascii',
                   'io.cmtsolution', 'io.cnv', 'io.css', 'io.csv',
                   'io.cybershake', 'io.dmx', 'io.focmec',
                   'io.hypodd', 'io.iaspei', 'io.gcf', 'io.gse2', 'io.json',
                   'io.kinemetrics', 'io.kml', 'io.mseed', 'io.ndk', 'io.nied',
                   'io.nlloc', 'io.nordic', 'io.pdas', 'io.pde', 'io.quakeml',
                   'io.reftek', 'io.rg16', 'io.sac', 'io.scardec', 'io.seg2',
                   'io.segy', 'io.seisan', 'io.sh', 'io.shapefile',
                   'io.seiscomp', 'io.stationtxt', 'io.stationxml', 'io.wav',
                   'io.win', 'io.xseed', 'io.y', 'io.zmap', 'realtime',
                   'scripts', 'signal', 'taup']
NETWORK_MODULES = ['clients.earthworm', 'clients.fdsn',
                   'clients.iris', 'clients.neic', 'clients.nrl',
                   'clients.seedlink', 'clients.syngine']
ALL_MODULES = DEFAULT_MODULES + NETWORK_MODULES

# default order of automatic format detection
WAVEFORM_PREFERRED_ORDER = ['MSEED', 'SAC', 'GSE2', 'SEISAN', 'SACXY', 'GSE1',
                            'Q', 'SH_ASC', 'SLIST', 'TSPAIR', 'Y', 'PICKLE',
                            'SEGY', 'SU', 'SEG2', 'WAV', 'WIN', 'CSS',
                            'NNSA_KB_CORE', 'AH', 'PDAS', 'KINEMETRICS_EVT',
                            'GCF', 'DMX', 'ALSEP_PSE', 'ALSEP_WTN',
                            'ALSEP_WTH', 'CYBERSHAKE']
EVENT_PREFERRED_ORDER = ['QUAKEML', 'SCML', 'NLLOC_HYP']
INVENTORY_PREFERRED_ORDER = ['STATIONXML', 'SEED', 'RESP', 'SCML']
# waveform plugins accepting a byteorder keyword
WAVEFORM_ACCEPT_BYTEORDER = ['MSEED', 'Q', 'SAC', 'SEGY', 'SU']

_sys_is_le = sys.byteorder == 'little'
NATIVE_BYTEORDER = _sys_is_le and '<' or '>'

# Define Obspy hard and soft dependencies
HARD_DEPENDENCIES = [
    "numpy", "scipy", "matplotlib", "lxml.etree", "setuptools",
    "sqlalchemy", "decorator", "requests"]
OPTIONAL_DEPENDENCIES = [
    "packaging", "pyproj", "pytest", "pytest-json-report", "pyshp",
    "geographiclib", "cartopy"]
DEPENDENCIES = HARD_DEPENDENCIES + OPTIONAL_DEPENDENCIES


class NamedTemporaryFile(io.BufferedIOBase):
    """
    Weak replacement for the Python's tempfile.TemporaryFile.

    This class is a replacement for :func:`tempfile.NamedTemporaryFile` but
    will work also with Windows 7/Vista's UAC.

    :type dir: str
    :param dir: If specified, the file will be created in that directory,
        otherwise the default directory for temporary files is used.
    :type suffix: str
    :param suffix: The temporary file name will end with that suffix. Defaults
        to ``'.tmp'``.

    .. rubric:: Example

    >>> with NamedTemporaryFile() as tf:
    ...     _ = tf.write(b"test")
    ...     os.path.exists(tf.name)
    True
    >>> # when using the with statement, the file is deleted at the end:
    >>> os.path.exists(tf.name)
    False

    >>> with NamedTemporaryFile() as tf:
    ...     filename = tf.name
    ...     with open(filename, 'wb') as fh:
    ...         _ = fh.write(b"just a test")
    ...     with open(filename, 'r') as fh:
    ...         print(fh.read())
    just a test
    >>> # when using the with statement, the file is deleted at the end:
    >>> os.path.exists(tf.name)
    False
    """

    def __init__(self, dir=None, suffix='.tmp', prefix='obspy-'):
        fd, self.name = tempfile.mkstemp(dir=dir, prefix=prefix, suffix=suffix)
        self._fileobj = os.fdopen(fd, 'w+b', 0)  # 0 -> do not buffer

    def read(self, *args, **kwargs):
        return self._fileobj.read(*args, **kwargs)

    def write(self, *args, **kwargs):
        return self._fileobj.write(*args, **kwargs)

    def seek(self, *args, **kwargs):
        self._fileobj.seek(*args, **kwargs)
        return self._fileobj.tell()

    def tell(self, *args, **kwargs):
        return self._fileobj.tell(*args, **kwargs)

    def close(self, *args, **kwargs):
        super(NamedTemporaryFile, self).close(*args, **kwargs)
        self._fileobj.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # @UnusedVariable
        self.close()
        Path(self.name).unlink()


def create_empty_data_chunk(delta, dtype, fill_value=None):
    """
    Creates an NumPy array depending on the given data type and fill value.

    If no ``fill_value`` is given a masked array will be returned.

    :param delta: Number of samples for data chunk
    :param dtype: NumPy dtype for returned data chunk
    :param fill_value: If ``None``, masked array is returned, else the
        array is filled with the corresponding value

    .. rubric:: Example

    >>> create_empty_data_chunk(3, 'int', 10)
    array([10, 10, 10])

    >>> create_empty_data_chunk(
    ...     3, 'f')  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    masked_array(data = [-- -- --],
                 mask = ...,
                 ...)
    """
    if fill_value is None:
        temp = np.ma.masked_all(delta, dtype=np.dtype(dtype))
        # fill with nan if float number and otherwise with a very small number
        if issubclass(temp.data.dtype.type, np.integer):
            temp.data[:] = np.iinfo(temp.data.dtype).min
        else:
            temp.data[:] = np.nan
    elif (isinstance(fill_value, list) or isinstance(fill_value, tuple)) \
            and len(fill_value) == 2:
        # if two values are supplied use these as samples bordering to our data
        # and interpolate between:
        ls = fill_value[0]
        rs = fill_value[1]
        # include left and right sample (delta + 2)
        interpolation = np.linspace(ls, rs, delta + 2)
        # cut ls and rs and ensure correct data type
        temp = np.require(interpolation[1:-1], dtype=np.dtype(dtype))
    else:
        temp = np.ones(delta, dtype=np.dtype(dtype))
        temp *= fill_value
    return temp


def get_example_file(filename):
    """
    Function to find the absolute path of a data file

    The ObsPy modules are installed to a custom installation directory.
    That is the path cannot be predicted. This functions searches for all
    installed ObsPy modules and checks whether the file is in any of
    the "tests/data/" or "data/" subdirectories.

    :param filename: A test file name to which the path should be returned.
    :return: Full path to file.

    .. rubric:: Example

    >>> get_example_file('slist.ascii')  # doctest: +SKIP
    /custom/path/to/obspy/io/ascii/tests/data/slist.ascii

    >>> get_example_file('does.not.exists')  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    OSError: Could not find file does.not.exists ...
    """
    for module in ALL_MODULES:
        try:
            mod = __import__("obspy.%s" % module,
                             fromlist=["obspy"])
        except ImportError:
            continue
        file_ = Path(mod.__path__[0]) / "tests" / "data" / filename
        if file_.is_file():
            return str(file_)
        file_ = Path(mod.__path__[0]) / "data" / filename
        if file_.is_file():
            return str(file_)
    msg = ("Could not find file %s in tests/data or data "
           "directory of ObsPy modules") % filename
    raise OSError(msg)


def get_entry_point_dist_name(entry_point):
    """
    Gets the distribution name from an entry point regardless of whether
    it comes from pkg_resources or importlib.metadata.

    :param entry_point: An EntryPoint object from pkg_resources
     or importlib.metadata
    :return: The distribution name as a string
    """
    # Check for pkg_resources style EntryPoint
    if hasattr(entry_point, 'dist') and hasattr(entry_point.dist, 'name'):
        return entry_point.dist.name

    # Check for importlib.metadata style EntryPoint
    if hasattr(entry_point, 'dist'):
        # Python 3.10+ approach where ep.dist might be the distribution name
        try:
            from importlib.metadata import distribution
            dist = distribution(entry_point.dist)
            return dist.metadata['Name']
        except KeyError:
            return str(entry_point.dist)

    # Fallback for older importlib.metadata versions
    import os
    import re

    if hasattr(entry_point, 'origin') and entry_point.origin:
        # Extract distribution name from the path
        path = entry_point.origin
        dist_info_dir = os.path.basename(os.path.dirname(path))
        match = re.match(r'(.+?)(?:-\d+.*)?.dist-info', dist_info_dir)
        if match:
            return match.group(1)

    # Last resort
    return entry_point.value.split('.')[0]


@functools.lru_cache(maxsize=None)
def _get_all_entry_points():
    """
    Get all entry points, using a cache to avoid repeated expensive
    scans of all installed python packages.

    Returns a flat list of all EntryPoint objects.
    """
    eps = importlib.metadata.entry_points()
    if not isinstance(eps, dict):
        # Python 3.10+: convert to list
        return list(eps)
    # Python 3.8/3.9: flatten the dict values to a list
    return [ep for group_eps in eps.values() for ep in group_eps]


@functools.lru_cache(maxsize=None)
def _get_entry_points(group, subgroup=None):
    """
    Gets a dictionary of all available plug-ins of a group or subgroup.

    :type group: str
    :param group: Group name.
    :type subgroup: str, optional
    :param subgroup: Subgroup name (defaults to None).
    :rtype: dict
    :returns: Dictionary of entry points of each plug-in.

    .. rubric:: Example

    >>> _get_entry_points(
    ...     'obspy.plugin.waveform') # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    {...'SLIST': EntryPoint(name='SLIST', value='obspy.io.ascii.core',
                            group='obspy.plugin.waveform')...}
    """
    if sys.version_info.minor < 10:
        # compatibility workaround for Python 3.8 and 3.9
        eps_all = _get_all_entry_points()

        # Build a dict grouped by group name for Python 3.8/3.9 compatibility
        eps_by_group = {}
        for ep in eps_all:
            if ep.group not in eps_by_group:
                eps_by_group[ep.group] = []
            eps_by_group[ep.group].append(ep)

        eps = eps_by_group.get(group, [])

        if subgroup:
            features = {}
            for ep in eps:
                # workaround to get the dist dict populated here
                if hasattr(ep, "dist"):
                    if not hasattr(ep.dist, "name"):
                        ep.dist.name = get_entry_point_dist_name(ep)
                else:
                    ep.dist = AttribDict({"name":
                                          get_entry_point_dist_name(ep)})
                subgroup_eps = eps_by_group.get(f'{group}.{ep.name}', [])
                for sub_ep in subgroup_eps:
                    if sub_ep.name == subgroup:
                        features[ep.name] = ep
                        break
        else:
            features = {}
            for ep in eps:
                # workaround to get the dist dict populated here
                if hasattr(ep, "dist"):
                    if not hasattr(ep.dist, "name"):
                        ep.dist.name = get_entry_point_dist_name(ep)
                else:
                    ep.dist = AttribDict({"name":
                                          get_entry_point_dist_name(ep)})
                features[ep.name] = ep
    else:
        # Efficient implementation: get all entry points once and filter in
        # memory instead of making multiple .select() calls which scan all
        # distributions and is very slow in Python 3.10+
        eps_all = _get_all_entry_points()

        # Filter to entries for the target group
        eps_for_group = [ep for ep in eps_all if ep.group == group]

        if subgroup:
            # Build a map of parent plugins in each subgroup
            # subgroups[parent_name] = [subgroup entries for that parent]
            subgroups = {}
            for ep in eps_all:
                # Check if this is a subgroup entry for our target group
                if ep.group.startswith(f'{group}.'):
                    parent_name = ep.group[len(group)+1:]  # e.g. "MSEED"
                    if parent_name not in subgroups:
                        subgroups[parent_name] = []
                    subgroups[parent_name].append(ep)

            # Only include parent plugins that have the matching subgroup
            features = {}
            for ep in eps_for_group:
                if (
                    ep.name in subgroups
                    and any(
                        sub_ep.name == subgroup
                        for sub_ep in subgroups[ep.name]
                    )
                ):
                    features[ep.name] = ep
        else:
            features = {ep.name: ep for ep in eps_for_group}
    return features


def _get_ordered_entry_points(group, subgroup=None, order_list=[]):
    """
    Gets a ordered dictionary of all available plug-ins of a group or subgroup.
    """
    # Copy so pop() below does not modify the lru_cache result.
    ep_dict = _get_entry_points(group, subgroup).copy()
    # loop through official supported waveform plug-ins and add them to
    # ordered dict of entry points
    entry_points = OrderedDict()
    for name in order_list:
        try:
            entry_points[name] = ep_dict.pop(name)
        except Exception:
            # skip plug-ins which are not installed
            continue
    # extend entry points with any left over waveform plug-ins
    entry_points.update(ep_dict)
    return entry_points


ENTRY_POINTS = {
    'trigger': _get_entry_points('obspy.plugin.trigger'),
    'filter': _get_entry_points('obspy.plugin.filter'),
    'rotate': _get_entry_points('obspy.plugin.rotate'),
    'detrend': _get_entry_points('obspy.plugin.detrend'),
    'interpolate': _get_entry_points('obspy.plugin.interpolate'),
    'integrate': _get_entry_points('obspy.plugin.integrate'),
    'differentiate': _get_entry_points('obspy.plugin.differentiate'),
    'waveform': _get_ordered_entry_points(
        'obspy.plugin.waveform', 'readFormat', WAVEFORM_PREFERRED_ORDER),
    'waveform_write': _get_ordered_entry_points(
        'obspy.plugin.waveform', 'writeFormat', WAVEFORM_PREFERRED_ORDER),
    'event': _get_ordered_entry_points('obspy.plugin.event', 'readFormat',
                                       EVENT_PREFERRED_ORDER),
    'event_write': _get_entry_points('obspy.plugin.event', 'writeFormat'),
    'taper': _get_entry_points('obspy.plugin.taper'),
    'inventory': _get_ordered_entry_points(
        'obspy.plugin.inventory', 'readFormat', INVENTORY_PREFERRED_ORDER),
    'inventory_write': _get_entry_points(
        'obspy.plugin.inventory', 'writeFormat'),
}


def _get_function_from_entry_point(group, type):
    """
    A "automagic" function searching a given dict of entry points for a valid
    entry point and returns the function call. Otherwise it will raise a
    default error message.

    .. rubric:: Example

    >>> _get_function_from_entry_point(
    ...     'detrend', 'simple')  # doctest: +ELLIPSIS
    <function simple at 0x...>

    >>> _get_function_from_entry_point('detrend', 'XXX')  # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    ValueError: Detrend type "XXX" is not supported. Supported types: ...
    """
    ep_dict = ENTRY_POINTS[group]
    try:
        # get entry point
        if type in ep_dict:
            entry_point = ep_dict[type]
        else:
            # search using lower cases only
            entry_point = [v for k, v in ep_dict.items()
                           if k.lower() == type.lower()][0]
    except (KeyError, IndexError):
        # check if any entry points are available at all
        if not ep_dict:
            msg = "Your current ObsPy installation does not support " + \
                  "any %s functions. Please make sure " + \
                  "SciPy is installed properly."
            raise ImportError(msg % (group.capitalize()))
        # ok we have entry points, but specified function is not supported
        msg = "%s type \"%s\" is not supported. Supported types: %s"
        raise ValueError(msg % (group.capitalize(), type, ', '.join(ep_dict)))
    # import function point
    # any issue during import of entry point should be raised, so the user has
    # a chance to correct the problem
    func = buffered_load_entry_point(entry_point.dist.name,
                                     'obspy.plugin.%s' % (group),
                                     entry_point.name)
    return func


def get_dependency_version(package_name, raw_string=False):
    """
    Get version information of a dependency package.

    :type package_name: str
    :param package_name: Name of package to return version info for
    :returns: Package version as a list of three integers or ``None`` if
        import fails. With option ``raw_string=True`` returns raw version
        string instead (or ``None`` if import fails).
        The last version number can indicate different things like it being a
        version from the old svn trunk, the latest git repo, some release
        candidate version, ...
        If the last number cannot be converted to an integer it will be set to
        0.
    """
    try:
        version_string = importlib.metadata.version(package_name)
    except importlib.metadata.PackageNotFoundError:
        return []
    if raw_string:
        return version_string
    version_list = version_string.split("rc")[0].strip("~")
    version_list = list(map(to_int_or_zero, version_list.split(".")))
    return version_list


NUMPY_VERSION = get_dependency_version('numpy')
SCIPY_VERSION = get_dependency_version('scipy')
MATPLOTLIB_VERSION = get_dependency_version('matplotlib')
CARTOPY_VERSION = get_dependency_version('cartopy')


def _read_from_plugin(plugin_type, filename, format=None, **kwargs):
    """
    Reads a single file from a plug-in's readFormat function.
    """
    # SC3ML plugin deprecated with 1.5.0 remove a few main releases later
    if format and format.upper() == 'SC3ML':
        warnings.warn(
            "Format 'SC3ML' is deprecated since ObsPy 1.5.0 and will be "
            "removed in a future release. Use 'SCML' instead.",
            category=ObsPyDeprecationWarning,
            stacklevel=3)
    if isinstance(filename, str):
        if not Path(filename).exists():
            msg = "[Errno 2] No such file or directory: '{}'".format(
                filename)
            raise FileNotFoundError(msg)
    eps = ENTRY_POINTS[plugin_type]
    # get format entry point
    format_ep = None
    if not format:
        # auto detect format - go through all known formats in given sort order
        for format_ep in eps.values():
            # search isFormat for given entry point
            is_format = buffered_load_entry_point(
                format_ep.dist.name,
                'obspy.plugin.%s.%s' % (plugin_type, format_ep.name),
                'isFormat')
            # If it is a file-like object, store the position and restore it
            # later to avoid that the isFormat() functions move the file
            # pointer.
            if hasattr(filename, "tell") and hasattr(filename, "seek"):
                position = filename.tell()
            else:
                position = None
            # check format
            is_format = is_format(filename)
            if position is not None:
                filename.seek(position, 0)
            if is_format:
                break
        else:
            raise TypeError('Unknown format for file %s' % filename)
    else:
        # format given via argument
        format = format.upper()
        try:
            format_ep = eps[format]
        except (KeyError, IndexError):
            msg = "Format \"%s\" is not supported. Supported types: %s"
            raise TypeError(msg % (format, ', '.join(eps)))
    # file format should be known by now
    try:
        # search readFormat for given entry point
        read_format = buffered_load_entry_point(
            format_ep.dist.name,
            'obspy.plugin.%s.%s' % (plugin_type, format_ep.name),
            'readFormat')
    except ImportError:
        msg = "Format \"%s\" is not supported. Supported types: %s"
        raise TypeError(msg % (format_ep.name, ', '.join(eps)))
    # read
    list_obj = read_format(filename, **kwargs)
    return list_obj, format_ep.name


def get_script_dir_name():
    """
    Get the directory of the current script file. This is more robust than
    using __file__.
    """
    return str(Path(inspect.getfile(
        inspect.currentframe())).parent.resolve())


def make_format_plugin_table(group="waveform", method="read", numspaces=4,
                             unindent_first_line=True):
    """
    Returns a markdown formatted table with read waveform plugins to insert
    in docstrings.

    >>> table = make_format_plugin_table("event", "write", 4, True)
    >>> print(table)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS +SKIP
    ======... ===========... ========================================...
    Format    Used Module    _`Linked Function Call`
    ======... ===========... ========================================...
    CMTSOLUTION  :mod:`...io.cmtsolution` :func:`..._write_cmtsolution`
    CNV       :mod:`...io.cnv`   :func:`obspy.io.cnv.core._write_cnv`
    HYPODDPHA :mod:`...io.hypodd`    :func:`obspy.io.hypodd.pha._write_pha`
    JSON      :mod:`...io.json`  :func:`obspy.io.json.core._write_json`
    KML       :mod:`obspy.io.kml` :func:`obspy.io.kml.core._write_kml`
    NLLOC_OBS :mod:`...io.nlloc` :func:`obspy.io.nlloc.core.write_nlloc_obs`
    NORDIC    :mod:`obspy.io.nordic` :func:`obspy.io.nordic.core.write_select`
    QUAKEML :mod:`...io.quakeml` :func:`obspy.io.quakeml.core._write_quakeml`
    SC3ML   :mod:`obspy.io.seiscomp`
                             :func:`obspy.io.seiscomp.event._write_scml`
    SCARDEC   :mod:`obspy.io.scardec`
                             :func:`obspy.io.scardec.core._write_scardec`
    SCML    :mod:`...io.seiscomp` :func:`obspy.io.seiscomp.event._write_scml`
    SHAPEFILE :mod:`obspy.io.shapefile`
                             :func:`obspy.io.shapefile.core._write_shapefile`
    ZMAP      :mod:`...io.zmap`  :func:`obspy.io.zmap.core._write_zmap`
    ======... ===========... ========================================...

    :type group: str
    :param group: Plugin group to search (e.g. "waveform" or "event").
    :type method: str
    :param method: Either 'read' or 'write' to select plugins based on either
        read or write capability.
    :type numspaces: int
    :param numspaces: Number of spaces prepended to each line (for indentation
        in docstrings).
    :type unindent_first_line: bool
    :param unindent_first_line: Determines if first line should start with
        prepended spaces or not.
    """
    method = method.lower()
    if method not in ("read", "write"):
        raise ValueError("no valid type: %s" % method)

    method = "%sFormat" % method
    eps = _get_ordered_entry_points("obspy.plugin.%s" % group, method,
                                    WAVEFORM_PREFERRED_ORDER)

    # Get all entry points once using module-level cache
    all_eps = _get_all_entry_points()

    mod_list = []
    for name, ep in eps.items():
        if sys.version_info.minor < 10:
            # compatibility workaround for Python 3.8 and 3.9
            module_short = ":mod:`%s`" % ".".join(ep.value.split(".")[:3])
            # Filter from cached list
            func_str = [
                _ep
                for _ep in all_eps
                if _ep.group == f'{ep.group}.{ep.name}' and _ep.name == method
            ][0].value
        else:
            module_short = ":mod:`%s`" % ".".join(ep.module.split(".")[:3])
            # Filter the pre-fetched all_eps instead of
            # calling entry_points() again
            matching_eps = [
                _ep
                for _ep in all_eps
                if _ep.group == f'{ep.group}.{ep.name}' and _ep.name == method
            ]
            if matching_eps:
                func_str = matching_eps[0].value
            else:
                # Fallback (shouldn't happen in normal operation)
                func_str = ep.value
        func_str = func_str.replace(':', '.')
        func_str = f':func:`{func_str}`'
        mod_list.append((name, module_short, func_str))

    mod_list = sorted(mod_list)
    headers = ["Format", "Used Module", "_`Linked Function Call`"]
    maxlens = [max([len(x[0]) for x in mod_list] + [len(headers[0])]),
               max([len(x[1]) for x in mod_list] + [len(headers[1])]),
               max([len(x[2]) for x in mod_list] + [len(headers[2])])]

    info_str = [" ".join(["=" * x for x in maxlens])]
    info_str.append(
        " ".join([headers[i].ljust(maxlens[i]) for i in range(3)]))
    info_str.append(info_str[0])

    for mod_infos in mod_list:
        info_str.append(
            " ".join([mod_infos[i].ljust(maxlens[i]) for i in range(3)]))
    info_str.append(info_str[0])

    ret = " " * numspaces + ("\n" + " " * numspaces).join(info_str)
    if unindent_first_line:
        ret = ret[numspaces:]
    return ret


def _add_format_plugin_table(func, group, method, numspaces=4):
    """
    A function to populate the docstring of func with its plugin table.
    """
    if func.__doc__ is not None and '%s' in func.__doc__:
        func.__doc__ = func.__doc__ % make_format_plugin_table(
            group, method, numspaces=numspaces)


class ComparingObject(object):
    """
    Simple base class that implements == and != based on self.__dict__
    """

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)


def _get_deprecated_argument_action(old_name, new_name, real_action='store'):
    """
    Specifies deprecated command-line arguments to scripts
    """
    message = '%s has been deprecated. Please use %s in the future.' % (
        old_name, new_name
    )

    from argparse import Action

    class _Action(Action):
        def __call__(self, parser, namespace, values, option_string=None):
            import warnings
            warnings.warn(message)

            # I wish there were an easier way...
            if real_action == 'store':
                setattr(namespace, self.dest, values)
            elif real_action == 'store_true':
                setattr(namespace, self.dest, True)
            elif real_action == 'store_false':
                setattr(namespace, self.dest, False)

    return _Action


def sanitize_filename(filename):
    """
    Adapted from Django's slugify functions.

    :param filename: The filename.
    """
    try:
        filename = filename.decode()
    except AttributeError:
        pass

    value = unicodedata.normalize('NFKD', filename).encode(
        'ascii', 'ignore').decode('ascii')
    # In constrast to django we allow dots and don't lowercase.
    value = re.sub(r'[^\w\.\s-]', '', value).strip()
    return re.sub(r'[-\s]+', '-', value)


def download_to_file(url, filename_or_buffer, chunk_size=1024):
    """
    Helper function to download a potentially large file.

    :param url: The URL to GET the data from.
    :type url: str
    :param filename_or_buffer: The filename_or_buffer or file-like object to
        download to.
    :type filename_or_buffer: str or file-like object
    :param chunk_size: The chunk size in bytes.
    :type chunk_size: int
    """
    import requests
    # Workaround for old request versions.
    try:
        r = requests.get(url, stream=True)
    except TypeError:
        r = requests.get(url)

    # Raise anything except for 200
    if r.status_code != 200:
        raise requests.HTTPError('%s HTTP Error: %s for url: %s'
                                 % (r.status_code, r.reason, url))

    if hasattr(filename_or_buffer, "write"):
        for chunk in r.iter_content(chunk_size=chunk_size):
            if not chunk:
                continue
            filename_or_buffer.write(chunk)
    else:
        with io.open(filename_or_buffer, "wb") as fh:
            for chunk in r.iter_content(chunk_size=chunk_size):
                if not chunk:
                    continue
                fh.write(chunk)


def _generic_reader(pathname_or_url=None, callback_func=None,
                    **kwargs):
    # convert pathlib.Path objects to str for compatibility.
    if isinstance(pathname_or_url, PurePath):
        pathname_or_url = str(pathname_or_url)
    if not isinstance(pathname_or_url, str):
        # first check if bytes
        if isinstance(pathname_or_url, bytes) and \
                pathname_or_url.strip().startswith(b'<'):
            # XML string
            return callback_func(io.BytesIO(pathname_or_url), **kwargs)
        # not a string OR bytes- we assume a file-like object
        try:
            # first try reading directly
            generic = callback_func(pathname_or_url, **kwargs)
        except TypeError:
            # if this fails, create a temporary file which is read directly
            # from the file system
            pathname_or_url.seek(0)
            with NamedTemporaryFile() as fh:
                fh.write(pathname_or_url.read())
                generic = callback_func(fh.name, **kwargs)
        return generic
    elif "://" in pathname_or_url[:10]:
        # URL
        # extract extension if any
        suffix = Path(Path(pathname_or_url).name).suffix
        if suffix == '':
            suffix = ".tmp"
        with NamedTemporaryFile(suffix=sanitize_filename(suffix)) as fh:
            download_to_file(url=pathname_or_url, filename_or_buffer=fh)
            generic = callback_func(fh.name, **kwargs)
        return generic
    else:
        pathname = pathname_or_url
        # File name(s)
        pathnames = sorted(glob.glob(pathname))
        if not pathnames:
            # try to give more specific information why the stream is empty
            if glob.has_magic(pathname) and not glob.glob(pathname):
                raise Exception("No file matching file pattern: %s" % pathname)
            elif not glob.has_magic(pathname) and not Path(pathname).is_file():
                raise IOError(2, "No such file or directory", pathname)

        generic = callback_func(pathnames[0], **kwargs)
        if len(pathnames) > 1:
            for filename in pathnames[1:]:
                generic.extend(callback_func(filename, **kwargs))
        return generic


def get_bytes_stream(file_or_stream):
    """
    Return a file-like object streaming bytes data (``bytes`` objects)

    :type file_or_stream: str, Path or file-like object
    :param file_or_stream: the input from which to return the stream of bytes
    :return: a file-like object streaming bytes data
    """
    if isinstance(file_or_stream, IOBase):
        return file_or_stream
    else:
        return open(file_or_stream, 'rb')


@contextmanager
def open_bytes_stream(file_or_stream):
    """
    Context manager to read bytes data stream from the argument,
    e.g.: ``with open_bytes_stream(file_or_stream)``

    :type file_or_stream: str, Path or file-like object
    :param file_or_stream: The input object to open.
        If this parameter is already a file-like object,
        it will not be closed on exit but the stream position
        will be reset so that the object can be reused
    :return: a context manager to read bytes data
    """
    cur_pos = None
    if isinstance(file_or_stream, IOBase):
        cur_pos = file_or_stream.tell()
    stream = get_bytes_stream(file_or_stream)
    try:
        yield stream
    finally:
        if cur_pos is None:  # str or Path passed: close stream
            stream.close()
        else:
            file_or_stream.seek(cur_pos, 0)


def get_text_stream(file_or_stream, encoding='utf-8'):
    """
    Return a file-like object streaming text data (``str`` objects)

    :type file_or_stream: str, Path or file-like object
    :param file_or_stream:  the input from which to return the stream of text
    :type encoding: str
    :param encoding: the encoding used. If the passed argument
        is already a file-like object of encoded text data, this
        argument is ignored
    :return: a file-like object streaming text data
    """
    if isinstance(file_or_stream, TextIOBase):
        return file_or_stream
    elif isinstance(file_or_stream, IOBase):
        return TextIOWrapper(file_or_stream, encoding=encoding)
    else:
        return open(file_or_stream, 'rt', encoding=encoding)


@contextmanager
def open_text_stream(file_or_stream, encoding='utf-8'):
    """
    Context manager releasing a text data stream from the argument,
    e.g.: ``with open_text_stream(file_or_stream)``

    :type file_or_stream: str, Path or file-like object.
    :param file_or_stream: The input object to open.
        If this parameter is already a file-like object,
        it will not be closed on exit but the stream position
        will be reset so that the object can be reused
    :type encoding: str
    :param encoding: the encoding used. If the passed argument
        is already a file-like object of encoded text data, this
        argument is ignored
    :return: a context manager to read text data
    """
    cur_pos = None
    if isinstance(file_or_stream, IOBase):
        cur_pos = file_or_stream.tell()
    stream = get_text_stream(file_or_stream, encoding=encoding)
    try:
        yield stream
    finally:
        if cur_pos is None:  # str or Path passed: close stream
            stream.close()
        else:
            # if we wrapped file_or_stream in a TextIOWrapper,
            # closing the latter would close also the underlying
            # stream. Simply detach the wrapper in this case:
            if isinstance(stream, TextIOWrapper) and \
                    stream is not file_or_stream:
                stream.detach()
            # reset position:
            file_or_stream.seek(cur_pos, 0)


class CatchAndAssertWarnings(warnings.catch_warnings):
    def __init__(self, clear=None, expected=None, show_all=True, **kwargs):
        """
        :type clear: list[str]
        :param clear: list of modules to clear warning
            registries on (e.g. ``["obspy.signal", "obspy.core"]``), in order
            to make sure any expected warnings will be shown and not suppressed
            because already raised in previously executed code.
        :type expected: list
        :param expected: list of 2-tuples specifying expected
            warnings that should be looked for when exiting the context
            manager. An ``AssertionError`` will be raised if any expected
            warning is not encountered. First item in tuple should be the
            class of the warning, second item should be a regex matching (a
            part of) the warning message (e.g.
            ``(ObsPyDeprecationWarning, 'Attribute .* is deprecated')``).
            Make sure to escape regex special characters like `(` or `.` with a
            backslash and provide message regex as a raw string.
        :type show_all: str
        :param show_all: Whether to set ``warnings.simplefilter('always')``
            when entering context.
        """
        self.registries_to_clear = clear
        self.expected_warnings = expected
        self.show_all = show_all
        # always record warnings, obviously..
        kwargs['record'] = True
        super(CatchAndAssertWarnings, self).__init__(**kwargs)

    def __enter__(self):
        self.warnings = super(CatchAndAssertWarnings, self).__enter__()
        if self.registries_to_clear:
            for modulename in self.registries_to_clear:
                self.clear_warning_registry(modulename)
        if self.show_all:
            warnings.simplefilter("always", Warning)
        # this will always return the list of warnings because we set
        # record=True
        return self.warnings

    def __exit__(self, *exc_info):
        super(CatchAndAssertWarnings, self).__exit__(self, *exc_info)
        # after cleanup, check expected warnings
        self._assert_warnings()

    @staticmethod
    def clear_warning_registry(modulename):
        """
        Clear warning registry of specified module

        :type modulename: str
        :param modulename: Full module name (e.g. ``'obspy.signal'``)
        """
        mod = importlib.import_module(modulename)
        try:
            registry = mod.__warningregistry__
        except AttributeError:
            pass
        else:
            registry.clear()

    def _assert_warnings(self):
        """
        Checks for expected warnings and raises an AssertionError if anyone of
        these is not encountered.
        """
        if not self.expected_warnings:
            return
        for category, regex in self.expected_warnings:
            for warning in self.warnings:
                if not isinstance(warning.message, category):
                    continue
                if not re.search(regex, str(warning.message)):
                    continue
                # found a matching warning, so break out
                break
            else:
                msg = 'Expected warning not raised: (%s, %s)'
                raise AssertionError(msg % (category.__name__, regex))


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# Author: Tom Eulenfeld
# Year: 2018
"""
SeismicHandler evt file bindings to ObsPy core module.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from collections import defaultdict
import io
from math import cos, pi
from warnings import warn

from obspy.core.event import (Arrival, Catalog, Event,
                              Magnitude, Origin, OriginQuality,
                              OriginUncertainty, Pick, ResourceIdentifier,
                              StationMagnitude, WaveformStreamID)
from obspy.core.event.header import EvaluationMode, EventType, PickOnset
from obspy.core.inventory.util import _resolve_seedid, _add_resolve_seedid_doc
from obspy.io.sh.core import to_utcdatetime


def _is_evt(filename):
    try:
        with open(filename, 'rb') as f:
            temp = f.read(20)
    except Exception:
        return False
    return b'Event ID' in temp


def _km2m(km):
    return 1000 * float(km)


def _km2deg(km):
    return float(km) / 111.195


def _event_type(et):
    if 'quake' in et:
        et = 'earthquake'
    return EventType(et)


MAG_MAP = {'ml': 'ML',
           'ms': 'MS',
           'mb': 'Mb',
           'mw': 'Mw'}


MAP = {
    'pick': {
        'onset time': ('time', to_utcdatetime),
        'phase flags': ('phase_hint', str),
        'onset type': ('onset', PickOnset),
        'pick type': ('evaluation_mode', EvaluationMode),
        'applied filter': ('filter_id', ResourceIdentifier),
        'sign': ('polarity', str)
    },
    'arrival': {
        'phase name': ('phase', str),
        'distance (deg)': ('distance', float),
        'theo. azimuth (deg)': ('azimuth', float),
        'weight': ('time_weight', float),
    },
    'origin': {
        'origin time': ('time', to_utcdatetime),
        'latitude': ('latitude', float),
        'longitude': ('longitude', float),
        'depth (km)': ('depth', _km2m),
        'error in origin time': ('time_errors', float),
        'error in latitude (km)': ('latitude_errors', _km2deg),
        'error in longitude (km)': ('longitude_errors', _km2deg),
        # (correction for lat in _read_evt)
        'error in depth (km)': ('depth_errors', _km2m),
        'no. of stations used': (
            'quality', lambda x: OriginQuality(used_station_count=int(x))),
        'source region': ('region', str)
    },
    'origin_uncertainty': {
        'error ellipse major': ('max_horizontal_uncertainty', _km2m),
        'error ellipse minor': ('min_horizontal_uncertainty', _km2m),
        'error ellipse strike': ('azimuth_max_horizontal_uncertainty', float)
    },
    'event': {
        'event type': ('event_type', _event_type)
    }
    # no dict for magnitudes, these are handled by function _mag
}


# define supported keys just for documentation
SUPPORTED_KEYS = ['event id', 'station code', 'component', 'magnitude (M?)',
                  'mean magnitude (M?)']
SUPPORTED_KEYS = sorted([key for obj in MAP.values() for key in obj] +
                        SUPPORTED_KEYS)


def _kw(obj, obj_name):
    kws = {}
    for source_key, (dest_key, convert) in MAP[obj_name].items():
        try:
            val = convert(obj[source_key])
        except KeyError:
            pass
        except ValueError as ex:
            warn(str(ex))
        else:
            kws[dest_key] = val
    return kws


def _mags(obj, evid, stamag=False, wid=None):
    mags = []
    pm = None
    for magtype1, magtype2 in MAG_MAP.items():
        magkey = 'mean ' * (not stamag) + 'magnitude ' + magtype1
        if magkey in obj:
            magv = obj[magkey]
            if 'inf' in magv:
                warn('invalid value for magnitude: %s (event id %s)'
                     % (magv, evid))
            else:
                magv = float(magv)
                mag = (StationMagnitude(station_magnitude_type=magtype2,
                                        mag=magv, waveform_id=wid)
                       if stamag else
                       Magnitude(magnitude_type=magtype2, mag=magv))
                mags.append(mag)
    if len(mags) == 1:
        pm = mags[0].resource_id
    return mags, pm


@_add_resolve_seedid_doc
def _read_evt(filename, encoding='utf-8', **kwargs):
    """
    Read a SeismicHandler EVT file and returns an ObsPy Catalog object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.event.read_events` function, call this
        instead.

    :param str encoding: encoding used (default: utf-8)

    :rtype: :class:`~obspy.core.event.Catalog`
    :return: An ObsPy Catalog object.

    .. note::
        The following fields are supported by this function: %s.

        Compare with http://www.seismic-handler.org/wiki/ShmDocFileEvt
    """
    with io.open(filename, 'r', encoding=encoding) as f:
        temp = f.read()
    # first create phases and phases_o dictionaries for different phases
    # and phases with origin information
    phases = defaultdict(list)
    phases_o = {}
    phase = {}
    evid = None
    for line in temp.splitlines():
        if 'End of Phase' in line:
            if 'origin time' in phase.keys():
                if evid in phases_o:
                    # found more than one origin
                    pass
                phases_o[evid] = phase
            phases[evid].append(phase)
            phase = {}
            evid = None
        elif line.strip() != '':
            try:
                key, value = line.split(':', 1)
            except ValueError:
                continue
            key = key.strip().lower()
            value = value.strip()
            if key == 'event id':
                evid = value
            elif value != '':
                phase[key] = value
    assert evid is None

    # now create obspy Events from phases and phases_o dictionaries
    events = []
    for evid in phases:
        picks = []
        arrivals = []
        stamags = []
        origins = []
        po = None
        magnitudes = []
        pm = None
        for p in phases[evid]:
            sta = p.get('station code', '')
            comp = p.get('component', '')
            pick_kwargs = _kw(p, 'pick')
            widargs = _resolve_seedid(
                    sta, comp, time=pick_kwargs['time'], **kwargs)
            wid = WaveformStreamID(*widargs)
            pick = Pick(waveform_id=wid, **pick_kwargs)
            arrival = Arrival(pick_id=pick.resource_id, **_kw(p, 'arrival'))
            picks.append(pick)
            arrivals.append(arrival)
            stamags_temp, _ = _mags(p, evid, stamag=True, wid=wid)
            stamags.extend(stamags_temp)
        if evid in phases_o:
            o = phases_o[evid]
            uncertainty = OriginUncertainty(**_kw(o, 'origin_uncertainty'))
            origin = Origin(arrivals=arrivals, origin_uncertainty=uncertainty,
                            **_kw(o, 'origin'))
            if origin.latitude is None or origin.longitude is None:
                warn('latitude or longitude not set for event %s' % evid)
            else:
                if origin.longitude_errors.uncertainty is not None:
                    origin.longitude_errors.uncertainty *= cos(
                        origin.latitude / 180 * pi)
                origins = [origin]
                po = origin.resource_id
            magnitudes, pm = _mags(o, evid)
        else:
            o = p
        event = Event(resource_id=ResourceIdentifier(evid),
                      picks=picks,
                      origins=origins,
                      magnitudes=magnitudes,
                      station_magnitudes=stamags,
                      preferred_origin_id=po,
                      preferred_magnitude_id=pm,
                      **_kw(o, 'event')
                      )
        events.append(event)
    return Catalog(events,
                   description='Created from SeismicHandler EVT format')


if _read_evt.__doc__ is not None and '%s' in _read_evt.__doc__:
    _read_evt.__doc__ = _read_evt.__doc__ % (SUPPORTED_KEYS,)



# -*- coding: utf-8 -*-
import warnings

from obspy.io.y.core import _is_y, _read_y


class TestCore():
    """
    Nanometrics Y file test suite.
    """
    def test_is_y_file(self, testdata):
        """
        Testing Y file format.
        """
        testfile = testdata['YAYT_BHZ_20021223.124800']
        assert _is_y(testfile)
        assert not _is_y("/path/to/slist.ascii")
        assert not _is_y("/path/to/tspair.ascii")

    def test_read_y_file(self, testdata):
        """
        Testing reading Y file format.
        """
        testfile = testdata['YAYT_BHZ_20021223.124800']
        st = _read_y(testfile)
        assert len(st) == 1
        tr = st[0]
        assert len(tr) == 18000
        assert tr.stats.sampling_rate == 100.0
        assert tr.stats.station == 'AYT'
        assert tr.stats.channel == 'BHZ'
        assert tr.stats.location == ''
        assert tr.stats.network == ''
        assert max(tr.data) == tr.stats.y.tag_series_info.max_amplitude
        assert min(tr.data) == tr.stats.y.tag_series_info.min_amplitude

    def test_ignore_non_ascii_tag_station_info(self, testdata):
        """
        Test faulty Y file containing non ASCII chars in TAG_STATION_INFO.
        """
        testfile = testdata['YAZRSPE.20100119.060433']
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            st = _read_y(testfile)
        assert len(w) == 1
        assert 'Invalid' in str(w[0])
        assert len(st) == 1
        tr = st[0]
        assert len(tr) == 16976
        assert tr.stats.sampling_rate == 50.0
        assert tr.stats.station == 'AZR'
        assert tr.stats.channel == 'E'
        assert tr.stats.location == 'SP'
        assert tr.stats.network == ''



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colorbar import ColorbarBase
from matplotlib.colors import Normalize

import obspy
from obspy.core.util import AttribDict
from obspy.imaging.cm import obspy_sequential
from obspy.signal.invsim import corn_freq_2_paz
from obspy.signal.array_analysis import array_processing


# Load data
st = obspy.read("https://examples.obspy.org/agfa.mseed")

# Set PAZ and coordinates for all 5 channels
st[0].stats.paz = AttribDict({
    'poles': [(-0.03736 - 0.03617j), (-0.03736 + 0.03617j)],
    'zeros': [0j, 0j],
    'sensitivity': 205479446.68601453,
    'gain': 1.0})
st[0].stats.coordinates = AttribDict({
    'latitude': 48.108589,
    'elevation': 0.450000,
    'longitude': 11.582967})

st[1].stats.paz = AttribDict({
    'poles': [(-0.03736 - 0.03617j), (-0.03736 + 0.03617j)],
    'zeros': [0j, 0j],
    'sensitivity': 205479446.68601453,
    'gain': 1.0})
st[1].stats.coordinates = AttribDict({
    'latitude': 48.108192,
    'elevation': 0.450000,
    'longitude': 11.583120})

st[2].stats.paz = AttribDict({
    'poles': [(-0.03736 - 0.03617j), (-0.03736 + 0.03617j)],
    'zeros': [0j, 0j],
    'sensitivity': 250000000.0,
    'gain': 1.0})
st[2].stats.coordinates = AttribDict({
    'latitude': 48.108692,
    'elevation': 0.450000,
    'longitude': 11.583414})

st[3].stats.paz = AttribDict({
    'poles': [(-4.39823 + 4.48709j), (-4.39823 - 4.48709j)],
    'zeros': [0j, 0j],
    'sensitivity': 222222228.10910088,
    'gain': 1.0})
st[3].stats.coordinates = AttribDict({
    'latitude': 48.108456,
    'elevation': 0.450000,
    'longitude': 11.583049})

st[4].stats.paz = AttribDict({
    'poles': [(-4.39823 + 4.48709j), (-4.39823 - 4.48709j), (-2.105 + 0j)],
    'zeros': [0j, 0j, 0j],
    'sensitivity': 222222228.10910088,
    'gain': 1.0})
st[4].stats.coordinates = AttribDict({
    'latitude': 48.108730,
    'elevation': 0.450000,
    'longitude': 11.583157})


# Instrument correction to 1Hz corner frequency
paz1hz = corn_freq_2_paz(1.0, damp=0.707)
st.simulate(paz_remove='self', paz_simulate=paz1hz)

# Execute array_processing
kwargs = dict(
    # slowness grid: X min, X max, Y min, Y max, Slow Step
    sll_x=-3.0, slm_x=3.0, sll_y=-3.0, slm_y=3.0, sl_s=0.03,
    # sliding window properties
    win_len=1.0, win_frac=0.05,
    # frequency properties
    frqlow=1.0, frqhigh=8.0, prewhiten=0,
    # restrict output
    semb_thres=-1e9, vel_thres=-1e9,
    stime=obspy.UTCDateTime("20080217110515"),
    etime=obspy.UTCDateTime("20080217110545")
)
out = array_processing(st, **kwargs)

# Plot

cmap = obspy_sequential

# make output human readable, adjust backazimuth to values between 0 and 360
t, rel_power, abs_power, baz, slow = out.T
baz[baz < 0.0] += 360

# choose number of fractions in plot (desirably 360 degree/N is an integer!)
N = 36
N2 = 30
abins = np.arange(N + 1) * 360. / N
sbins = np.linspace(0, 3, N2 + 1)

# sum rel power in bins given by abins and sbins
hist, baz_edges, sl_edges = \
    np.histogram2d(baz, slow, bins=[abins, sbins], weights=rel_power)

# transform to radian
baz_edges = np.radians(baz_edges)

# add polar and colorbar axes
fig = plt.figure(figsize=(8, 8))
cax = fig.add_axes([0.85, 0.2, 0.05, 0.5])
ax = fig.add_axes([0.10, 0.1, 0.70, 0.7], polar=True)
ax.set_theta_direction(-1)
ax.set_theta_zero_location("N")

dh = abs(sl_edges[1] - sl_edges[0])
dw = abs(baz_edges[1] - baz_edges[0])

# circle through backazimuth
for i, row in enumerate(hist):
    bars = ax.bar((i * dw) * np.ones(N2),
                  height=dh * np.ones(N2),
                  width=dw, bottom=dh * np.arange(N2),
                  color=cmap(row / hist.max()))

ax.set_xticks(np.linspace(0, 2 * np.pi, 4, endpoint=False))
ax.set_xticklabels(['N', 'E', 'S', 'W'])

# set slowness limits
ax.set_ylim(0, 3)
[i.set_color('grey') for i in ax.get_yticklabels()]
ColorbarBase(cax, cmap=cmap,
             norm=Normalize(vmin=hist.min(), vmax=hist.max()))

plt.show()



# -*- coding: utf-8 -*-
"""
Earthworm Wave Server client for ObsPy.

:copyright:
    The ObsPy Development Team (devs@obspy.org) & Victor Kress
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

.. seealso:: http://www.isti2.com/ew/PROGRAMMER/wsv_protocol.html
"""
from fnmatch import fnmatch

from obspy import Stream, UTCDateTime
from obspy.clients.earthworm.waveserver import get_menu, read_wave_server_v


class Client(object):
    """
    A Earthworm Wave Server client.

    :type host: str
    :param host: Host name of the remote Earthworm Wave Server server.
    :type port: int
    :param port: Port of the remote Earthworm Wave Server server.
    :type timeout: int, optional
    :param timeout: Seconds before a connection timeout is raised (default is
        ``None``).
    :type debug: bool, optional
    :param debug: Enables verbose output of the connection handling (default is
        ``False``).
    """
    def __init__(self, host, port, timeout=None, debug=False):
        """
        Initializes a Earthworm Wave Server client.

        See :class:`obspy.clients.earthworm.client.Client` for all parameters.
        """
        self.host = host
        self.port = port
        self.timeout = timeout
        self.debug = debug

    def get_waveforms(self, network, station, location, channel, starttime,
                      endtime, cleanup=True):
        """
        Retrieves waveform data from Earthworm Wave Server and returns an ObsPy
        Stream object.

        :type filename: str
        :param filename: Name of the output file.
        :type network: str
        :param network: Network code, e.g. ``'UW'``.
        :type station: str
        :param station: Station code, e.g. ``'TUCA'``.
        :type location: str
        :param location: Location code, e.g. ``'--'``.
        :type channel: str
        :param channel: Channel code, e.g. ``'BHZ'``. Last character (i.e.
            component) can be a wildcard ('?' or '*') to fetch `Z`, `N` and
            `E` component.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start date and time.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End date and time.
        :return: ObsPy :class:`~obspy.core.stream.Stream` object.
        :type cleanup: bool
        :param cleanup: Specifies whether perfectly aligned traces should be
            merged or not. See :meth:`obspy.core.stream.Stream.merge` for
            ``method=-1``.

        .. rubric:: Example

        >>> from obspy.clients.earthworm import Client
        >>> client = Client("pubavo1.wr.usgs.gov", 16022)
        >>> dt = UTCDateTime() - 15000  # now - 15000 seconds
        >>> st = client.get_waveforms('AV', 'AKV', '', 'BHE', dt, dt + 10)
        >>> st.plot()  # doctest: +SKIP
        >>> st = client.get_waveforms('AV', 'AKV', '', 'BH*', dt, dt + 10)
        >>> st.plot()  # doctest: +SKIP

        .. plot::

            from obspy.clients.earthworm import Client
            from obspy import UTCDateTime
            client = Client("pubavo1.wr.usgs.gov", 16022, timeout=5)
            dt = UTCDateTime() - 15000  # now - 15000 seconds
            st = client.get_waveforms('AV', 'AKV', '', 'BHE', dt, dt + 10)
            st.plot()
            st = client.get_waveforms('AV', 'AKV', '', 'BH*', dt, dt + 10)
            st.plot()
        """
        # replace wildcards in last char of channel and fetch all 3 components
        if channel[-1] in "?*":
            st = Stream()
            for comp in ("Z", "N", "E"):
                channel_new = channel[:-1] + comp
                st += self.get_waveforms(network, station, location,
                                         channel_new, starttime, endtime,
                                         cleanup=cleanup)
            return st
        if location == '':
            location = '--'
        scnl = (station, channel, network, location)
        # fetch waveform
        tbl = read_wave_server_v(self.host, self.port, scnl, starttime,
                                 endtime, timeout=self.timeout,
                                 cleanup=cleanup)
        # create new stream
        st = Stream()
        for tb in tbl:
            st.append(tb.get_obspy_trace())
        if cleanup:
            st._cleanup()
        st.trim(starttime, endtime)
        return st

    def save_waveforms(self, filename, network, station, location, channel,
                       starttime, endtime, format="MSEED", cleanup=True):
        """
        Writes a retrieved waveform directly into a file.

        :type filename: str
        :param filename: Name of the output file.
        :type network: str
        :param network: Network code, e.g. ``'UW'``.
        :type station: str
        :param station: Station code, e.g. ``'TUCA'``.
        :type location: str
        :param location: Location code, e.g. ``''``.
        :type channel: str
        :param channel: Channel code, e.g. ``'BHZ'``. Last character (i.e.
            component) can be a wildcard ('?' or '*') to fetch `Z`, `N` and
            `E` component.
        :type starttime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param starttime: Start date and time.
        :type endtime: :class:`~obspy.core.utcdatetime.UTCDateTime`
        :param endtime: End date and time.
        :type format: str, optional
        :param format: Output format. One of ``"MSEED"``, ``"GSE2"``,
            ``"SAC"``, ``"SACXY"``, ``"Q"``, ``"SH_ASC"``, ``"SEGY"``,
            ``"SU"``, ``"WAV"``. See the Supported Formats section in method
            :meth:`~obspy.core.stream.Stream.write` for a full list of
            supported formats. Defaults to ``'MSEED'``.
        :type cleanup: bool
        :param cleanup: Specifies whether perfectly aligned traces should be
            merged or not. See :meth:`~obspy.core.stream.Stream.merge`,
            `method` -1 or :meth:`~obspy.core.stream.Stream._cleanup`.
        :return: None

        .. rubric:: Example

        >>> from obspy.clients.earthworm import Client
        >>> client = Client("pubavo1.wr.usgs.gov", 16022)
        >>> t = UTCDateTime() - 2000  # now - 2000 seconds
        >>> client.save_waveforms('AV.AKV.--.BHE.mseed',
        ...                       'AV', 'AKV', '', 'BHE',
        ...                       t, t + 10, format='MSEED')  # doctest: +SKIP
        """
        st = self.get_waveforms(network, station, location, channel, starttime,
                                endtime, cleanup=cleanup)
        st.write(filename, format=format)

    def get_availability(self, network="*", station="*", location="*",
                         channel="*"):
        """
        Gets a list of data available on the server.

        This method returns information about what time series data is
        available on the server. The query can optionally be restricted to
        specific network, station, channel and/or location criteria.

        :type network: str
        :param network: Network code, e.g. ``'UW'``, wildcards allowed.
        :type station: str
        :param station: Station code, e.g. ``'TUCA'``, wildcards allowed.
        :type location: str
        :param location: Location code, e.g. ``'--'``, wildcards allowed.
        :type channel: str
        :param channel: Channel code, e.g. ``'BHZ'``, wildcards allowed.
        :rtype: list
        :return: List of tuples with information on the available data. One
            tuple consists of network, station, location, channel
            (all strings), start time and end time
            (both as :class:`~obspy.core.utcdatetime.UTCDateTime`).

        .. rubric:: Example

        >>> from obspy.clients.earthworm import Client
        >>> client = Client("pubavo1.wr.usgs.gov", 16022, timeout=5)
        >>> response = client.get_availability(
        ...     network="AV", station="AKV", channel="BH*")
        >>> print(response)  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        [('AV',
          'AKV',
          '--',
          'BHE',
          UTCDateTime(...),
          UTCDateTime(...)),
         ('AV',
          'AKV',
          '--',
          'BHN',
          UTCDateTime(...),
          UTCDateTime(...)),
         ('AV',
          'AKV',
          '--',
          'BHZ',
          UTCDateTime(...),
          UTCDateTime(...))]
        """
        # build up possibly wildcarded trace id pattern for query
        if location == '':
            location = '--'
        pattern = ".".join((network, station, location, channel))
        # get overview of all available data, winston wave servers can not
        # restrict the query via network, station etc. so we do that manually
        response = get_menu(self.host, self.port, timeout=self.timeout)
        # reorder items and convert time info to UTCDateTime
        response = [(x[3], x[1], x[4], x[2], UTCDateTime(x[5]),
                     UTCDateTime(x[6])) for x in response]
        # restrict results according to user input
        response = [x for x in response if fnmatch(".".join(x[:4]), pattern)]
        return response


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
SAC module helper functions and data.

"""
import sys
import warnings

import numpy as np

from obspy import UTCDateTime
from obspy.core import Stats

from . import header as HD  # noqa

# ------------- DATA ----------------------------------------------------------
TWO_DIGIT_YEAR_MSG = ("SAC file with 2-digit year header field encountered. "
                      "This is not supported by the SAC file format standard. "
                      "Prepending '19'.")


# ------------- SAC-SPECIFIC EXCEPTIONS ---------------------------------------
class SacError(Exception):
    """
    Raised if the SAC file is corrupt or if necessary information
    in the SAC file is missing.
    """
    pass


class SacIOError(SacError, IOError):
    """
    Raised if the given SAC file can't be read.
    """
    pass


class SacInvalidContentError(SacError):
    """
    Raised if headers and/or data are not valid.
    """
    pass


class SacHeaderError(SacError):
    """
    Raised if header has issues.
    """
    pass


class SacHeaderTimeError(SacHeaderError, ValueError):
    """
    Raised if header has invalid "nz" times.
    """
    pass


# ------------- VALIDITY CHECKS -----------------------------------------------
def is_valid_enum_str(hdr, name):
    # is this a valid string name for this hdr
    # assume that, if a value isn't in HD.ACCEPTED_VALS, it's not valid
    if hdr in HD.ACCEPTED_VALS:
        tf = name in HD.ACCEPTED_VALS[hdr]
    else:
        tf = False
    return tf


def is_valid_enum_int(hdr, val, allow_null=True):
    # is this a valid integer for this hdr.
    if hdr in HD.ACCEPTED_VALS:
        accep = [HD.ENUM_VALS[nm] for nm in HD.ACCEPTED_VALS[hdr]]
        if allow_null:
            accep += [HD.INULL]
        tf = val in accep
    else:
        tf = False
    return tf


# ------------- GENERAL -------------------------------------------------------
def _convert_enum(header, converter, accep):
    # header : dict, SAC header
    # converter : dict, {source value: target value}
    # accep : dict, {header name: acceptable value list}

    # TODO: use functools.partial/wraps?
    for hdr, val in header.items():
        if hdr in HD.ACCEPTED_VALS:
            if val in accep[hdr]:
                header[hdr] = converter[val]
            else:
                msg = 'Unrecognized enumerated value "{}" for header "{}"'
                raise ValueError(msg.format(val, hdr))

    return header


def enum_string_to_int(header):
    """
    Convert enumerated string values in header dictionary to int values.
    """
    out = _convert_enum(header, converter=HD.ENUM_VALS, accep=HD.ACCEPTED_VALS)
    return out


def enum_int_to_string(header):
    """
    Convert enumerated int values in header dictionary to string values.
    """
    out = _convert_enum(header, converter=HD.ENUM_NAMES, accep=HD.ACCEPTED_INT)
    return out


def byteswap(*arrays):
    """
    Swapping of bytes for provided arrays.

    **Notes**

    arr.newbyteorder('S') swaps dtype interpretation, but not bytes in memory
    arr.byteswap() swaps bytes in memory, but not dtype interpretation
    arr.byteswap(True).newbyteorder('S') completely swaps both

    .. seealso::
        https://docs.scipy.org/doc/numpy/user/basics.byteswapping.html
    """
    return [arr.view(arr.dtype.newbyteorder('S')) for arr in arrays]


def is_same_byteorder(bo1, bo2):
    """
    Deal with all the ways to compare byte order string representations.

    :param bo1: Byte order string. Can be one of {'l', 'little', 'L', '<',
        'b', 'big', 'B', '>', 'n', 'native','N', '='}
    :type bo1: str
    :param bo2: Byte order string. Can be one of {'l', 'little', 'L', '<',
        'b', 'big', 'B', '>', 'n', 'native','N', '='}
    :type bo1: str

    :rtype: bool
    :return: True of same byte order.
    """
    # TODO: extend this as is_same_byteorder(*byteorders) using itertools
    be = ('b', 'big', '>')
    le = ('l', 'little', '<')
    ne = ('n', 'native', '=')
    ok = be + le + ne

    if (bo1.lower() not in ok) or (bo2.lower() not in ok):
        raise ValueError("Unrecognized byte order string.")

    # make native decide what it is
    bo1 = sys.byteorder if bo1.lower() in ne else bo1
    bo2 = sys.byteorder if bo2.lower() in ne else bo2

    return (bo1.lower() in le) == (bo2.lower() in le)


def _clean_str(value, strip_whitespace=True):
    """
    Remove null values and whitespace, return a str

    This fn is used in two places: in SACTrace.read, to sanitize strings for
    SACTrace, and in sac_to_obspy_header, to sanitize strings for making a
    Trace that the user may have manually added.
    """
    null_term = value.find('\x00')
    if null_term >= 0:
        value = value[:null_term] + " " * len(value[null_term:])

    if strip_whitespace:
        value = value.strip()

    return value


# TODO: do this in SACTrace?
def sac_to_obspy_header(sacheader, round_sampling_interval=True):
    """
    Make an ObsPy Stats header dictionary from a SAC header dictionary.

    :param sacheader: SAC header dictionary.
    :type sacheader: dict
    :param round_sampling_interval: Whether to round sampling interval to
        microseconds before calculating sampling rate to avoid floating point
        accuracy issues with some SAC files (see #3408)
    :type round_sampling_interval: bool

    :rtype: :class:`~obspy.core.trace.Stats`
    :return: Filled ObsPy Stats header.
    """

    # 1. get required sac header values
    try:
        npts = sacheader['npts']
        delta = sacheader['delta']
    except KeyError:
        msg = "Incomplete SAC header information to build an ObsPy header."
        raise KeyError(msg)

    assert npts != HD.INULL
    assert delta != HD.FNULL
    #
    # 2. get time
    try:
        reftime = get_sac_reftime(sacheader)
    except (SacError, ValueError, TypeError):
        # ObsPy doesn't require a valid reftime
        reftime = UTCDateTime(0.0)

    b = sacheader.get('b', HD.FNULL)
    #
    # 3. get optional sac header values
    calib = sacheader.get('scale', HD.FNULL)
    kcmpnm = sacheader.get('kcmpnm', HD.SNULL)
    kstnm = sacheader.get('kstnm', HD.SNULL)
    knetwk = sacheader.get('knetwk', HD.SNULL)
    khole = sacheader.get('khole', HD.SNULL)
    #
    # 4. deal with null values
    b = b if (b != HD.FNULL) else 0.0
    calib = calib if (calib != HD.FNULL) else 1.0
    kcmpnm = kcmpnm if (kcmpnm != HD.SNULL) else ''
    kstnm = kstnm if (kstnm != HD.SNULL) else ''
    knetwk = knetwk if (knetwk != HD.SNULL) else ''
    khole = khole if (khole != HD.SNULL) else ''
    #
    # 5. transform to obspy values
    # nothing is null
    stats = {}
    stats['npts'] = npts
    # sampling rate can have floating point issues due to it being stored as a
    # single precision floating point sampling interval in seconds, see #3408
    # usually it should be ok potentially, but it seems that some SAC data
    # sources additionally don't have the sampling interval encoded as the
    # closest possible binary representation but rather the closest lesser
    # binary representation (e.g. b'\x0b\xd7#=' for 0.04) when the next
    # possible higher number would be closer to the "exact" value (here
    # b'\n\xd7#=') which seems to be the tipping point of the floating point
    # inaccuracy making it through our internal conversion to double precision
    # sampling rate. here, we check if the conventional conversion and the
    # "safer" way of rounding off (to microseconds in sample spacing) lead to a
    # difference and warn about the fact, since the rounding approach actually
    # could introduce buggy results for real weird unaligned sampling rates
    sample_spacing_conventional = np.float32(delta)
    sample_spacing_rounded_off = round(np.float64(delta), 6)
    sampling_rate_conventional = np.float32(1.) / np.float32(delta)
    sampling_rate_rounded_off = 1.0 / round(np.float64(delta), 6)
    if round_sampling_interval:
        if sampling_rate_conventional != sampling_rate_rounded_off:
            msg = (f"Sample spacing read from SAC file "
                   f"({sample_spacing_conventional:.9f} when rounded to "
                   f"nanoseconds) was rounded of to microsecond precision "
                   f"({sample_spacing_rounded_off:.9f}) to avoid floating "
                   f"point issues when converting to sampling rate (see "
                   f"#3408)")
            warnings.warn(msg)
        stats['sampling_rate'] = sampling_rate_rounded_off
    else:
        stats['sampling_rate'] = sampling_rate_conventional
    stats['network'] = _clean_str(knetwk)
    stats['station'] = _clean_str(kstnm)
    stats['channel'] = _clean_str(kcmpnm)
    stats['location'] = _clean_str(khole)
    stats['calib'] = calib

    # store _all_ provided SAC header values
    stats['sac'] = sacheader.copy()

    # get first sample absolute time as UTCDateTime
    # always add the begin time (if it's defined) to get the given
    # SAC reference time, no matter which iztype is given
    # b may be non-zero, even for iztype 'ib', especially if it was used to
    #   store microseconds from obspy_to_sac_header
    stats['starttime'] = UTCDateTime(reftime) + b

    return Stats(stats)


def split_microseconds(microseconds):
    # Returns milliseconds and remainder microseconds
    milliseconds = microseconds // 1000
    microseconds = (microseconds - milliseconds * 1000)

    return milliseconds, microseconds


def utcdatetime_to_sac_nztimes(utcdt):
    # Returns a dict of integer nz-times and remainder microseconds
    nztimes = {}
    nztimes['nzyear'] = utcdt.year
    nztimes['nzjday'] = utcdt.julday
    nztimes['nzhour'] = utcdt.hour
    nztimes['nzmin'] = utcdt.minute
    nztimes['nzsec'] = utcdt.second
    # nz times don't have enough precision, so push microseconds into b,
    # using integer arithmetic
    millisecond, microsecond = split_microseconds(utcdt.microsecond)
    nztimes['nzmsec'] = millisecond

    return nztimes, microsecond


def obspy_to_sac_header(stats, keep_sac_header=True):
    """
    Merge a primary with a secondary header, reconciling some differences.

    :param stats: Filled ObsPy Stats header
    :type stats: dict or :class:`~obspy.core.trace.Stats`
    :param keep_sac_header: If keep_sac_header is True, old stats.sac
        header values are kept, and a minimal set of values are updated from
        the stats dictionary according to these guidelines:

        * npts, delta always come from stats
        * If a valid old reftime is found, the new b and e will be made
          and properly referenced to it. All other old SAC headers are simply
          carried along.
        * If the old SAC reftime is invalid and relative time headers are set,
          a SacHeaderError exception will be raised.
        * If the old SAC reftime is invalid, no relative time headers are set,
          and "b" is set, "e" is updated from stats and other old SAC headers
          are carried along.
        * If the old SAC reftime is invalid, no relative time headers are set,
          and "b" is not set, the reftime will be set from stats.starttime
          (with micro/milliseconds precision adjustments) and "b" and "e" are
          set accordingly.
        * If 'kstnm', 'knetwk', 'kcmpnm', or 'khole' are not set or differ
          from Stats values 'station', 'network', 'channel', or 'location',
          they are taken from the Stats values.

        If keep_sac_header is False, a new SAC header is constructed from only
        information found in the Stats dictionary, with some other default
        values introduced.  It will be an iztype 9 ("ib") file, with small
        reference time adjustments for micro/milliseconds precision issues.
        SAC headers nvhdr, level, lovrok, and iftype are always produced.
    :type keep_sac_header: bool
    :rtype merged: dict
    :return: SAC header
    """
    header = {}
    oldsac = stats.get('sac', {})

    if keep_sac_header and oldsac:
        header.update(oldsac)

        try:
            reftime = get_sac_reftime(header)
        except SacHeaderTimeError:
            reftime = None

        relhdrs = [hdr for hdr in HD.RELHDRS
                   if header.get(hdr) not in (None, HD.FNULL)]

        if reftime:
            # Set current 'b' relative to the old reftime.
            b = stats['starttime'] - reftime
        else:
            # Invalid reference time. Relative times like 'b' cannot be
            # unambiguously referenced to stats.starttime.
            if 'b' in relhdrs:
                # Assume no trimming/expanding of the Trace occurred relative
                # to the old 'b', and just use the old 'b' value.
                b = header['b']
            else:
                # Assume it's an iztype=ib (9) type file. Also set iztype?
                b = 0

            # Set the stats.starttime as the reftime and set 'b' and 'e'.
            # ObsPy issue 1204
            reftime = stats['starttime'] - b
            nztimes, microsecond = utcdatetime_to_sac_nztimes(reftime)
            header.update(nztimes)
            b += (microsecond * 1e-6)

        header['b'] = b
        header['e'] = b + (stats['endtime'] - stats['starttime'])

        # Merge some values from stats if they're missing in the SAC header
        # ObsPy issues 1204, 1457
        # XXX: If Stats values are empty/"" and SAC header values are real,
        #   this will replace the real SAC values with SAC null values.
        for sachdr, statshdr in [('kstnm', 'station'), ('knetwk', 'network'),
                                 ('kcmpnm', 'channel'), ('khole', 'location')]:
            if (header.get(sachdr) in (None, HD.SNULL)) or \
               (header.get(sachdr).strip() != stats[statshdr]):
                header[sachdr] = stats[statshdr] or HD.SNULL
    else:
        # SAC header from Stats only.

        # Here, set headers from Stats that would otherwise depend on the old
        # SAC header
        header['iztype'] = 9
        starttime = stats['starttime']
        # nz times don't have enough precision, so push microseconds into b,
        # using integer arithmetic
        nztimes, microsecond = utcdatetime_to_sac_nztimes(starttime)
        header.update(nztimes)

        header['b'] = microsecond * 1e-6

        # we now have correct b, npts, delta, and nz times
        header['e'] = header['b'] + (stats['npts'] - 1) * stats['delta']

        header['scale'] = stats.get('calib', HD.FNULL)

        # NOTE: overwrites existing SAC headers
        # nulls for these are '', which stats.get(hdr, HD.SNULL) won't catch
        header['kcmpnm'] = stats['channel'] if stats['channel'] else HD.SNULL
        header['kstnm'] = stats['station'] if stats['station'] else HD.SNULL
        header['knetwk'] = stats['network'] if stats['network'] else HD.SNULL
        header['khole'] = stats['location'] if stats['location'] else HD.SNULL

        header['lpspol'] = True
        header['lcalda'] = False

    # ObsPy issue 1204
    header['nvhdr'] = 6
    header['leven'] = 1
    header['lovrok'] = 1
    header['iftype'] = 1

    # ObsPy issue #1317
    header['npts'] = stats['npts']
    header['delta'] = stats['delta']

    return header


def get_sac_reftime(header):
    """
    Get SAC header reference time as a UTCDateTime instance from a SAC header
    dictionary.

    Builds the reference time from SAC "nz" time fields. Raises
    :class:`SacHeaderTimeError` if any time fields are null.

    :param header: SAC header
    :type header: dict

    :rtype: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :returns: SAC reference time.
    """
    # NOTE: epoch seconds can be got by:
    # (reftime - datetime.datetime(1970,1,1)).total_seconds()
    try:
        yr = header['nzyear']
        nzjday = header['nzjday']
        nzhour = header['nzhour']
        nzmin = header['nzmin']
        nzsec = header['nzsec']
        nzmsec = header['nzmsec']
    except KeyError as e:
        # header doesn't have all the keys
        msg = "Not enough time information: {}".format(e)
        raise SacHeaderTimeError(msg)

    if 0 <= yr <= 99:
        warnings.warn(TWO_DIGIT_YEAR_MSG)
        yr += 1900

    try:
        reftime = UTCDateTime(year=yr, julday=nzjday, hour=nzhour,
                              minute=nzmin, second=nzsec,
                              microsecond=nzmsec * 1000)
    except (ValueError, TypeError):
        msg = "Invalid time headers. May contain null values."
        raise SacHeaderTimeError(msg)

    return reftime



# -*- coding: utf-8 -*-
"""
HypoDD PHA read support.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import io
from math import cos
from numpy import deg2rad
from warnings import warn

from obspy import UTCDateTime
from obspy.core.event import (
    Catalog, Event, Origin, Magnitude, Pick, WaveformStreamID, Arrival,
    OriginQuality)
from obspy.core.inventory.util import (
    _add_resolve_seedid_doc, _add_resolve_seedid_ph2comp_doc, _resolve_seedid)


DEG2KM = 111.2


def _block2event(block, eventid_map, **kwargs):
    """
    Read HypoDD event block
    """
    lines = block.strip().splitlines()
    yr, mo, dy, hr, mn, sc, la, lo, dp, mg, eh, ez, rms, id_ = lines[0].split()
    if eventid_map is not None and id_ in eventid_map:
        id_ = eventid_map[id_]
    time = UTCDateTime(int(yr), int(mo), int(dy), int(hr), int(mn), float(sc),
                       strict=False)
    laterr = None if float(eh) == 0 else float(eh) / DEG2KM
    lonerr = (None if laterr is None or float(la) > 89 else
              laterr / cos(deg2rad(float(la))))
    ez = None if float(ez) == 0 else float(ez) * 1000
    rms = None if float(rms) == 0 else float(rms)
    picks = []
    arrivals = []
    for line in lines[1:]:
        sta, reltime, weight, phase = line.split()
        widargs = _resolve_seedid(sta, '', time=time, phase=phase, **kwargs)
        wid = WaveformStreamID(*widargs)
        pick = Pick(waveform_id=wid, phase_hint=phase,
                    time=time + float(reltime))
        arrival = Arrival(phase=phase, pick_id=pick.resource_id,
                          time_weight=float(weight))
        picks.append(pick)
        arrivals.append(arrival)
    qu = OriginQuality(associated_phase_count=len(picks), standard_error=rms)
    origin = Origin(arrivals=arrivals,
                    resource_id="smi:local/origin/" + id_,
                    quality=qu,
                    latitude=float(la),
                    longitude=float(lo),
                    depth=1000 * float(dp),
                    latitude_errors=laterr,
                    longitude_errors=lonerr,
                    depth_errors=ez,
                    time=time)
    if mg.lower() == 'nan':
        magnitudes = []
        preferred_magnitude_id = None
    else:
        magnitude = Magnitude(mag=mg, resource_id="smi:local/magnitude/" + id_)
        magnitudes = [magnitude]
        preferred_magnitude_id = magnitude.resource_id
    event = Event(resource_id="smi:local/event/" + id_,
                  picks=picks,
                  origins=[origin],
                  magnitudes=magnitudes,
                  preferred_origin_id=origin.resource_id,
                  preferred_magnitude_id=preferred_magnitude_id)
    return event


def _is_pha(filename):
    try:
        with open(filename, 'rb') as f:
            line = f.readline()
        assert line.startswith(b'#')
        assert len(line.split()) == 15
        yr, mo, dy, hr, mn, sc = line.split()[1:7]
        UTCDateTime(int(yr), int(mo), int(dy), int(hr), int(mn), float(sc),
                    strict=False)
    except Exception:
        return False
    else:
        return True


@_add_resolve_seedid_ph2comp_doc
@_add_resolve_seedid_doc
def _read_pha(filename, eventid_map=None, encoding='utf-8', **kwargs):
    """
    Read a HypoDD PHA file and returns an ObsPy Catalog object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.event.read_events` function, call this
        instead.

    :param str filename: File or file-like object in text mode.
    :param dict ph2comp: mapping of phases to components
        (default: {'P': 'Z', 'S': 'N'})
    :param dict eventid_map: Desired mapping of hypodd event ids (dict values)
        to event resource ids (dict keys).
        The returned dictionary of the HYPODDPHA writing operation can be used.
        By default, ids are not mapped.
    :param str encoding: encoding used (default: utf-8)

    :rtype: :class:`~obspy.core.event.Catalog`
    :return: An ObsPy Catalog object.
    """
    if eventid_map is not None:
        eventid_map = {v: k for k, v in eventid_map.items()}
    with io.open(filename, 'r', encoding=encoding) as f:
        text = f.read()
    events = [_block2event(block, eventid_map, **kwargs)
              for block in text.split('#')[1:]]
    return Catalog(events)


PHA1 = ('# {o.time.year} {o.time.month:>2} {o.time.day:>2} {o.time.hour:>2}'
        ' {o.time.minute:>2} {o.time.second:>2}.{o.time.microsecond:06d}  '
        '{o.latitude} {o.longitude} {depth}  {mag}  {he} {ve} {rms}'
        '   {evid:>9}\n')
PHA2 = '{p.waveform_id.station_code:6}  {relt:.4f}  {weight}  {p.phase_hint}\n'


def _map_eventid(evid, eventid_map, used_ids, counter):
    idpha = evid
    if evid in eventid_map:
        idpha = eventid_map[evid]
        if not idpha.isdigit() or len(idpha) > 9:
            msg = ('Invalid value in eventid_map, pha event id has to be '
                   'digit with max 9 digits')
            raise ValueError(msg)
        return idpha
    if not idpha.isdigit():
        idpha = ''.join(char for char in idpha if char.isdigit())
    if len(idpha) > 9:
        idpha = idpha[:9]
    while idpha == '' or idpha in used_ids:
        idpha = str(counter[0])
        counter[0] += 1
    if idpha != evid:
        eventid_map[evid] = idpha
    used_ids.add(idpha)
    return idpha


def _write_pha(catalog, filename, eventid_map=None,
               **kwargs):  # @UnusedVariable
    """
    Write a HypoDD PHA file.

    .. warning::
        This function should NOT be called directly, it registers via the
        the :meth:`~obspy.core.event.Catalog.write` method of an
        ObsPy :class:`~obspy.core.event.Catalog` object, call this instead.

    :type catalog: :class:`~obspy.core.event.catalog.Catalog`
    :param catalog: The ObsPy Catalog object to write.
    :type filename: str or file-like object
    :param filename: Filename to write or open file-like object.
    :param dict eventid_map: Desired mapping of event resource ids (dict keys)
        to hypodd event ids (dict values).
        HYPODD expects integer event ids with maximal 9 digits. If the event
        resource id is not present in the mapping,
        the event resource id is stripped of all non-digit characters and
        truncated to a length of 9 chars. If this method does not generate a
        valid hypodd event id, a counter starting at 1000 is used.

    :returns: Dictionary eventid_map with mapping of event resource id to
        hypodd event id. Items are only present if both ids are different.
    """
    if len(catalog) >= 10**10:
        warn('Writing a very large catalog will use event ids that might not '
             'be readable by HypoDD.')
    lines = []
    if eventid_map is None:
        eventid_map = {}
    args_map_eventid = (eventid_map, set(eventid_map.values()), [1])
    for event in catalog:
        try:
            ori = event.preferred_origin() or event.origins[0]
        except IndexError:
            warn(f'Skipping writing event with missing origin: {event}')
            continue
        try:
            mag = event.preferred_magnitude() or event.magnitudes[0]
        except IndexError:
            warn('Missing magnitude will be set to 0.0')
            mag = 0.
        else:
            mag = mag.mag
        evid = event.resource_id.id
        evid = evid.split('/')[-1] if '/' in evid else evid
        evid = _map_eventid(evid, *args_map_eventid)
        rms = (ori.quality.standard_error if 'quality' in ori and ori.quality
               else None)
        rms = rms if rms is not None else 0.0
        he1 = ori.latitude_errors.uncertainty if ori.latitude_errors else None
        he2 = (ori.longitude_errors.uncertainty if ori.longitude_errors
               else None)
        shortening = cos(deg2rad(ori.latitude))
        he = max(0. if he1 is None else he1 * DEG2KM,
                 0. if he2 is None else he2 * DEG2KM * shortening)
        ve = ori.depth_errors.uncertainty if ori.depth_errors else None
        ve = 0. if ve is None else ve / 1000
        line = PHA1.format(o=ori, depth=ori.depth / 1000, mag=mag,
                           he=he, ve=ve, rms=rms,
                           evid=evid)
        lines.append(line)
        weights = {str(arrival.pick_id): arrival.time_weight
                   for arrival in ori.arrivals if arrival.time_weight}
        for pick in event.picks:
            weight = weights.get(str(pick.resource_id), 1.)
            line = PHA2.format(p=pick, relt=pick.time - ori.time,
                               weight=weight)
            lines.append(line)
    data = ''.join(lines)
    try:
        with open(filename, 'w') as fh:
            fh.write(data)
    except TypeError:
        filename.write(data)
    return None if len(eventid_map) == 0 else eventid_map


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
The hoctavbands.core test suite.
"""
import numpy as np
import pytest
from scipy import signal

from obspy.signal import hoctavbands, util


# only tests for windowed data are implemented currently

class TestHoctavbands():
    """
    Test cases for half octav bands
    """
    @pytest.fixture(scope="function", autouse=True)
    def setup_data(self, testdata):
        self.res = np.loadtxt(testdata['3cssan.hy.1.MBGA_Z'])
        data = np.loadtxt(testdata['MBGA_Z.ASC'])
        # self.path = os.path.dirname(__file__)
        # self.res = np.loadtxt("3cssan.hy.1.MBGA_Z")
        # data = np.loadtxt("MBGA_Z.ASC")
        self.n = 256
        self.fs = 75
        self.smoothie = 3
        self.fk = [2, 1, 0, -1, -2]
        self.inc = int(0.05 * self.fs)
        self.fc1 = 0.68
        self.nofb = 8
        # [0] Time (k*inc)
        # [1] A_norm
        # [2] dA_norm
        # [3] dAsum
        # [4] dA2sum
        # [5] ct
        # [6] dct
        # [7] omega
        # [8] domega
        # [9] sigma
        # [10] dsigma
        # [11] log_cepstrum
        # [12] log_cepstrum
        # [13] log_cepstrum
        # [14] dperiod
        # [15] ddperiod
        # [16] bandwidth
        # [17] dbwith
        # [18] cfreq
        # [19] dcfreq
        # [20] hob1
        # [21] hob2
        # [22] hob3
        # [23] hob4
        # [24] hob5
        # [25] hob6
        # [26] hob7
        # [27] hob8
        # [28] phi12
        # [29] dphi12
        # [30] phi13
        # [31] dphi13
        # [32] phi23
        # [33] dphi23
        # [34] lv_h1
        # [35] lv_h2
        # [36] lv_h3
        # [37] dlv_h1
        # [38] dlv_h2
        # [39] dlv_h3
        # [40] rect
        # [41] drect
        # [42] plan
        # [43] dplan
        self.data_win, self.nwin, self.no_win = \
            util.enframe(data, signal.windows.hamming(self.n), self.inc)

    def test_hoctavbands(self):
        """
        """
        hob = hoctavbands.sonogram(self.data_win, self.fs, self.fc1,
                                   self.nofb, self.no_win)
        rms = np.sqrt(np.sum((hob[:, 0] - self.res[:, 20]) ** 2) /
                      np.sum(self.res[:, 20] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((hob[:, 1] - self.res[:, 21]) ** 2) /
                      np.sum(self.res[:, 21] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((hob[:, 2] - self.res[:, 22]) ** 2) /
                      np.sum(self.res[:, 22] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((hob[:, 3] - self.res[:, 23]) ** 2) /
                      np.sum(self.res[:, 23] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((hob[:, 4] - self.res[:, 24]) ** 2) /
                      np.sum(self.res[:, 24] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((hob[:, 5] - self.res[:, 25]) ** 2) /
                      np.sum(self.res[:, 25] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((hob[:, 6] - self.res[:, 26]) ** 2) /
                      np.sum(self.res[:, 26] ** 2))
        assert rms < 1.0e-5
        rms = np.sqrt(np.sum((hob[:, 7] - self.res[:, 27]) ** 2) /
                      np.sum(self.res[:, 27] ** 2))
        assert rms < 1.0e-5



# -*- coding: utf-8 -*-
"""
Convenience class for iterating over miniSEED records in a file.

Version 2017.114
"""
import ctypes as C  # NOQA

from obspy import UTCDateTime
from obspy.io.mseed.headers import (HPTMODULUS, MS_NOERROR, MS_ENDOFFILE,
                                    MSFileParam, MSRecord, clibmseed)


class _MSRIterator(object):
    """
    Class for iterating through miniSEED records in a file.

    :ivar msr: MSRecord (pointer to)
    :ivar msf: MSFileparam (pointer to)
    :ivar file: filename
    :ivar offset: Current offset

    :param filename: File to read
    :param startoffset: Offset in bytes to start reading the file
    :param reclen: If reclen is 0 the length of the first record is auto-
        detected. All subsequent records are then expected to have the
        same record length. If reclen is negative the length of every
        record is automatically detected. Defaults to -1.
    :param dataflag: Controls whether data samples are unpacked, defaults
        to False.
    :param skipnotdata: If true (not zero) any data chunks read that to do
        not have valid data record indicators will be skipped. Defaults to
        True (1).
    :param verbose: Controls verbosity from 0 to 2. Defaults to None (0).

    .. rubric:: Notes

    The elements of the MSRecord struct are available through `contents`
    on the pointer to the MSrecord, e.g. `_MSRIterator.msr.contents.reclen`.

    The raw record as a byte array is available at:
    `_MSRIterator.msr.contents.record`
    and can be used with Python I/O routines using:
    `ctypes.string_at(_MSRIterator.msr.contents.record,
    _MSRIterator.msr.contents.reclen))

    .. rubric:: Example

    from msriterator import _MSRIterator

    mseedfile = "test.mseed"

    for msri in _MSRIterator(filename=mseedfile, dataflag=False):

        print ("{:d}: {}, reclen: {}, samples: {}, starttime: {}, endtime: {}".
               format(msri.get_offset(),
                      msri.get_srcname(quality=False),
                      msri.msr.contents.reclen,
                      msri.msr.contents.samplecnt,
                      msri.get_starttime(),
                      msri.get_endtime()))
    """

    def __init__(self, filename, startoffset=0,
                 reclen=-1, dataflag=0, skipnotdata=1, verbose=0,
                 raise_errors=True):

        self.msr = clibmseed.msr_init(C.POINTER(MSRecord)())
        self.msf = C.POINTER(MSFileParam)()  # NULL pointer
        self.file = filename

        # Best guess at off_t type
        self._c_off_t = C.c_long

        self.fpos = self._c_off_t(0)
        if startoffset != 0:
            self.fpos = self._c_off_t(-startoffset)

        self.reclen = reclen
        self.dataflag = 1 if dataflag else 0
        self.skipnotdata = 1 if skipnotdata else 0
        self.verbose = verbose
        self.raise_errors = raise_errors

    def __iter__(self):
        return self

    def __next__(self):
        """
        Read next record from file.
        """
        errcode = clibmseed.ms_readmsr_r(C.pointer(self.msf),
                                         C.pointer(self.msr),
                                         self.file.encode('ascii', 'strict'),
                                         self.reclen,
                                         C.byref(self.fpos),
                                         None,
                                         self.skipnotdata,
                                         self.dataflag,
                                         self.verbose)

        if errcode == MS_ENDOFFILE:
            raise StopIteration()

        if self.raise_errors:
            if errcode != MS_NOERROR:
                raise Exception("Error %d in ms_readmsr_r" % errcode)

        return self

    def __del__(self):
        """
        Method for deallocating MSFileParam and MSRecord structures.
        """
        errcode = clibmseed.ms_readmsr_r(C.pointer(self.msf),
                                         C.pointer(self.msr),
                                         None, -1, None, None, 0, 0, 0)
        if errcode != MS_NOERROR:
            raise Exception("Error %d in ms_readmsr_r" % errcode)

    def get_srcname(self, quality=False):
        """
        Return record start time
        """
        srcname = C.create_string_buffer(50)

        clibmseed.msr_srcname.argtypes = [C.POINTER(MSRecord),
                                          C.POINTER(C.c_char),
                                          C.c_int]
        clibmseed.msr_srcname.restype = C.POINTER(C.c_char)

        if quality:
            clibmseed.msr_srcname(self.msr, srcname, 1)
        else:
            clibmseed.msr_srcname(self.msr, srcname, 0)

        return srcname.value.decode('utf-8')

    def get_starttime(self):
        """
        Return record start time as UTCDateTime
        """
        hptime = self.msr.contents.starttime
        return UTCDateTime(hptime / HPTMODULUS)

    def get_endtime(self):
        """
        Return record end time as UTCDateTime
        """
        hptime = clibmseed.msr_endtime(self.msr)
        return UTCDateTime(hptime / HPTMODULUS)

    def get_startepoch(self):
        """
        Return record start time as epoch time
        """
        return self.msr.contents.starttime / HPTMODULUS

    def get_endepoch(self):
        """
        Return record end time as epoch time
        """
        hptime = clibmseed.msr_endtime(self.msr)
        return hptime / HPTMODULUS

    def set_offset(self, value):
        """
        Set file reading position
        """
        self.fpos = self._c_off_t(-value)

    def get_offset(self):
        """
        Return offset into file for current record
        """
        return self.fpos.value

    offset = property(get_offset, set_offset)


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
SH bindings to ObsPy core module.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import io
from pathlib import Path
import numpy as np

from obspy import Stream, Trace, UTCDateTime
from obspy.core import Stats
from obspy.core.compatibility import from_buffer
from obspy.core.util import open_bytes_stream


MONTHS = ['JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN', 'JUL', 'AUG', 'SEP',
          'OCT', 'NOV', 'DEC']

MONTHS_DE = ['JAN', 'FEB', 'MAR', 'APR', 'MAI', 'JUN', 'JUL', 'AUG', 'SEP',
             'OKT', 'NOV', 'DEZ']

SH_IDX = {
    'LENGTH': 'L001',
    'SIGN': 'I011',
    'EVENTNO': 'I012',
    'MARK': 'I014',
    'DELTA': 'R000',
    'CALIB': 'R026',
    'DISTANCE': 'R011',
    'AZIMUTH': 'R012',
    'SLOWNESS': 'R018',
    'INCI': 'R013',
    'DEPTH': 'R014',
    'MAGNITUDE': 'R015',
    'LAT': 'R016',
    'LON': 'R017',
    'SIGNOISE': 'R022',
    'PWDW': 'R023',
    'DCVREG': 'R024',
    'DCVINCI': 'R025',
    'COMMENT': 'S000',
    'STATION': 'S001',
    'OPINFO': 'S002',
    'FILTER': 'S011',
    'QUAL': 'S012',
    'COMP': 'C000',
    'CHAN1': 'C001',
    'CHAN2': 'C002',
    'BYTEORDER': 'C003',
    'START': 'S021',
    'P-ONSET': 'S022',
    'S-ONSET': 'S023',
    'ORIGIN': 'S024'
}

STANDARD_ASC_HEADERS = ['START', 'COMP', 'CHAN1', 'CHAN2', 'STATION', 'CALIB']

SH_KEYS_INT = [k for (k, v) in SH_IDX.items() if v.startswith('I')]
SH_KEYS_FLOAT = [k for (k, v) in SH_IDX.items() if v.startswith('R')]
INVERTED_SH_IDX = {v: k for k, v in SH_IDX.items()}


def _is_asc(filename):
    """
    Checks whether a file is a Seismic Handler ASCII file or not.

    :type filename: str
    :param filename: Name of the ASCII file to be checked.
    :rtype: bool
    :return: ``True`` if a Seismic Handler ASCII file.

    .. rubric:: Example

    >>> _is_asc("/path/to/QFILE-TEST-ASC.ASC")  #doctest: +SKIP
    True
    """
    # first six chars should contain 'DELTA:'
    try:
        with open_bytes_stream(filename) as f:
            temp = f.read(6)
    except Exception:
        return False
    if temp != b'DELTA:':
        return False
    return True


def _read_asc(filename, headonly=False, skip=0, delta=None, length=None,
              **kwargs):  # @UnusedVariable
    """
    Reads a Seismic Handler ASCII file and returns an ObsPy Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :type filename: str or StringIO
    :param filename: ASCII file to be read.
    :type headonly: bool, optional
    :param headonly: If set to True, read only the head. This is most useful
        for scanning available data in huge (temporary) data sets.
    :type skip: int, optional
    :param skip: Number of lines to be skipped from top of file. If defined
        only one trace is read from file.
    :type delta: float, optional
    :param delta: If ``skip`` is used, ``delta`` defines sample offset in
        seconds.
    :type length: int, optional
    :param length: If ``skip`` is used, ``length`` defines the number of values
        to be read.
    :rtype: :class:`~obspy.core.stream.Stream`
    :return: A ObsPy Stream object.

    .. rubric:: Example

    >>> from obspy import read
    >>> st = read("/path/to/QFILE-TEST-ASC.ASC")
    >>> st  # doctest: +ELLIPSIS
    <obspy.core.stream.Stream object at 0x...>
    >>> print(st)  # doctest: +ELLIPSIS
    3 Trace(s) in Stream:
    .TEST..BHN | 2009-10-01T12:46:01.000000Z - ... | 20.0 Hz, 801 samples
    .TEST..BHE | 2009-10-01T12:46:01.000000Z - ... | 20.0 Hz, 801 samples
    .WET..HHZ  | 2010-01-01T01:01:05.999000Z - ... | 100.0 Hz, 4001 samples
    """
    # read file and split text into channels
    channels = []
    headers = {}
    data = io.StringIO()
    with open_bytes_stream(filename) as fh:
        for line in fh.readlines()[skip:]:
            if isinstance(line, bytes):
                line = line.decode('ascii')
            if line.isspace():
                # blank line
                # check if any data fetched yet
                if len(headers) == 0 and data.tell() == 0:
                    continue
                # append current channel
                data.seek(0)
                channels.append((headers, data))
                # create new channel
                headers = {}
                data = io.StringIO()
                if skip:
                    # if skip is set only one trace is read, everything else
                    # makes no sense.
                    break
                continue
            elif line[0].isalpha():
                # header entry
                key, value = line.split(':', 1)
                key = key.strip()
                value = value.strip()
                headers[key] = value
            elif not headonly:
                # data entry - may be written in multiple columns
                data.write(line.strip() + ' ')
    # create ObsPy stream object
    stream = Stream()
    # custom header
    custom_header = {}
    if delta:
        custom_header["delta"] = delta
    if length:
        custom_header["npts"] = length

    for headers, data in channels:
        # create Stats
        header = Stats(custom_header)
        header['sh'] = {}
        channel = [' ', ' ', ' ']
        # generate headers
        for key, value in headers.items():
            if key == 'DELTA':
                header['delta'] = float(value)
            elif key == 'LENGTH':
                header['npts'] = int(value)
            elif key == 'CALIB':
                header['calib'] = float(value)
            elif key == 'STATION':
                header['station'] = value
            elif key == 'COMP':
                channel[2] = value[0]
            elif key == 'CHAN1':
                channel[0] = value[0]
            elif key == 'CHAN2':
                channel[1] = value[0]
            elif key == 'START':
                # 01-JAN-2009_01:01:01.0
                # 1-OCT-2009_12:46:01.000
                header['starttime'] = to_utcdatetime(value)
            else:
                # everything else gets stored into sh entry
                if key in SH_KEYS_INT:
                    header['sh'][key] = int(value)
                elif key in SH_KEYS_FLOAT:
                    header['sh'][key] = float(value)
                else:
                    header['sh'][key] = value
        # set channel code
        header['channel'] = ''.join(channel)
        if headonly:
            # skip data
            stream.append(Trace(header=header))
        else:
            # read data
            data = np.loadtxt(data, dtype=np.float32, ndmin=1)

            # cut data if requested
            if skip and length:
                data = data[:length]

            # use correct value in any case
            header["npts"] = len(data)

            stream.append(Trace(data=data, header=header))
    return stream


def _write_asc(stream, filename, included_headers=None, npl=4,
               custom_format="%-.6e", append=False,
               **kwargs):  # @UnusedVariable
    """
    Writes a Seismic Handler ASCII file from given ObsPy Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        the :meth:`~obspy.core.stream.Stream.write` method of an
        ObsPy :class:`~obspy.core.stream.Stream` object, call this instead.

    :type stream: :class:`~obspy.core.stream.Stream`
    :param stream: The ObsPy Stream object to write.
    :type filename: str
    :param filename: Name of the ASCII file to write.
    :type npl: int, optional
    :param npl: Number of data columns in file, default to four.
    :type included_headers: list or None, optional
    :param included_headers: If set to a list, only these header entries will
        be written to file. DELTA and LENGTH are written in any case. If it's
        set to None, a basic set will be included.
    :type custom_format: str, optional
    :param custom_format: Parameter for number formatting of samples, defaults
        to "%-.6e".
    :type append: bool, optional
    :param append: If filename exists append all data to file, default False.
    """
    if included_headers is None:
        included_headers = STANDARD_ASC_HEADERS

    sio = io.StringIO()
    for trace in stream:
        # write headers
        sio.write("DELTA: %-.6e\n" % (trace.stats.delta))
        sio.write("LENGTH: %d\n" % trace.stats.npts)
        # additional headers
        for key, value in trace.stats.get('sh', {}).items():
            if included_headers and key not in included_headers:
                continue
            sio.write("%s: %s\n" % (key, value))
        # special format for start time
        if "START" in included_headers:
            dt = trace.stats.starttime
            sio.write("START: %s\n" % from_utcdatetime(dt))
        # component must be split
        if len(trace.stats.channel) > 2 and "COMP" in included_headers:
            sio.write("COMP: %c\n" % trace.stats.channel[2])
        if len(trace.stats.channel) > 0 and "CHAN1" in included_headers:
            sio.write("CHAN1: %c\n" % trace.stats.channel[0])
        if len(trace.stats.channel) > 1 and "CHAN2" in included_headers:
            sio.write("CHAN2: %c\n" % trace.stats.channel[1])
        if "STATION" in included_headers:
            sio.write("STATION: %s\n" % trace.stats.station)
        if "CALIB" in included_headers:
            sio.write("CALIB: %-.6e\n" % (trace.stats.calib))
        # write data in npl columns
        mask = ([''] * (npl - 1)) + ['\n']
        delimiter = mask * ((trace.stats.npts // npl) + 1)
        delimiter = delimiter[:trace.stats.npts - 1]
        delimiter.append('\n')
        for (sample, delim) in zip(trace.data, delimiter):
            value = custom_format % (sample)
            sio.write("%s %s" % (value, delim))
        sio.write("\n")
    if append:
        mode = 'ab'
    else:
        mode = 'wb'
    with open(filename, mode=mode) as fh:
        sio.seek(0)
        fh.write(sio.read().encode('ascii', 'strict'))


def _is_q(filename):
    """
    Checks whether a file is a Seismic Handler Q file or not.

    :type filename: str
    :param filename: Name of the Q file to be checked.
    :rtype: bool
    :return: ``True`` if a Seismic Handler Q file.

    .. rubric:: Example

    >>> _is_q("/path/to/QFILE-TEST.QHD")  #doctest: +SKIP
    True
    """
    # file must start with magic number 43981
    try:
        with open(filename, 'rb') as f:
            temp = f.read(5)
    except Exception:
        return False
    if temp != b'43981':
        return False
    return True


def _read_q(filename, headonly=False, data_directory=None, byteorder='=',
            **kwargs):  # @UnusedVariable
    """
    Reads a Seismic Handler Q file and returns an ObsPy Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    :type filename: str
    :param filename: Q header file to be read. Must have a `QHD` file
        extension.
    :type headonly: bool, optional
    :param headonly: If set to True, read only the head. This is most useful
        for scanning available data in huge (temporary) data sets.
    :type data_directory: str, optional
    :param data_directory: Data directory where the corresponding QBN file can
        be found.
    :type byteorder: str, optional
    :param byteorder: Enforce byte order for data file. This is important for
        Q files written in older versions of Seismic Handler, which don't
        explicit state the `BYTEORDER` flag within the header file. Can be
        little endian (``'<'``), big endian (``'>'``), or native byte order
        (``'='``). Defaults to ``'='``.
    :rtype: :class:`~obspy.core.stream.Stream`
    :return: A ObsPy Stream object.

    Q files consists of two files per data set:

    * a ASCII header file with file extension `QHD` and the
    * binary data file with file extension `QBN`.

    The read method only accepts header files for the ``filename`` parameter.
    ObsPy assumes that the corresponding data file is within the same directory
    if the ``data_directory`` parameter is not set. Otherwise it will search
    in the given ``data_directory`` for a file with the `QBN` file extension.
    This function should NOT be called directly, it registers via the
    ObsPy :func:`~obspy.core.stream.read` function, call this instead.

    .. rubric:: Example

    >>> from obspy import read
    >>> st = read("/path/to/QFILE-TEST.QHD")
    >>> st    #doctest: +ELLIPSIS
    <obspy.core.stream.Stream object at 0x...>
    >>> print(st)  # doctest: +ELLIPSIS
    3 Trace(s) in Stream:
    .TEST..BHN | 2009-10-01T12:46:01.000000Z - ... | 20.0 Hz, 801 samples
    .TEST..BHE | 2009-10-01T12:46:01.000000Z - ... | 20.0 Hz, 801 samples
    .WET..HHZ  | 2010-01-01T01:01:05.999000Z - ... | 100.0 Hz, 4001 samples
    """
    if not headonly:
        if not data_directory:
            path = Path(filename)
            data_file = Path(path.parent) / (path.stem+".QBN")

        else:
            path = Path(filename)
            data_file = Path(data_directory) / Path(filename).stem+".QBN"
        if not Path(data_file).is_file():
            msg = "Can't find corresponding QBN file at %s."
            raise IOError(msg % data_file)
        fh_data = open(data_file, 'rb')
    # loop through read header file
    with open(filename, 'rt') as fh:
        lines = fh.read().splitlines()
    # number of comment lines
    cmtlines = int(lines[0][5:7])
    # trace lines
    traces = {}
    i = -1
    id = ''
    for line in lines[cmtlines:]:
        cid = int(line[0:2])
        if cid != id:
            id = cid
            i += 1
        traces.setdefault(i, '')
        traces[i] += line[3:]
    # create stream object
    stream = Stream()
    for id in sorted(traces.keys()):
        # fetch headers
        header = {}
        header['sh'] = {
            "FROMQ": True,
            "FILE": Path(filename).stem,
        }
        channel = ['', '', '']
        npts = 0
        for item in traces[id].split('~'):
            key = item.lstrip()[0:4]
            value = item.lstrip()[5:]
            if key == 'L001':
                npts = header['npts'] = int(value)
            elif key == 'L000':
                continue
            elif key == 'R000':
                header['delta'] = float(value)
            elif key == 'R026':
                header['calib'] = float(value)
            elif key == 'S001':
                header['station'] = value
            elif key == 'C000' and value:
                channel[2] = value[0]
            elif key == 'C001' and value:
                channel[0] = value[0]
            elif key == 'C002' and value:
                channel[1] = value[0]
            elif key == 'C003':
                if value == '<' or value == '>':
                    byteorder = header['sh']['BYTEORDER'] = value
            elif key == 'S021':
                # 01-JAN-2009_01:01:01.0
                # 1-OCT-2009_12:46:01.000
                header['starttime'] = to_utcdatetime(value)
            elif key == 'S022':
                header['sh']['P-ONSET'] = to_utcdatetime(value)
            elif key == 'S023':
                header['sh']['S-ONSET'] = to_utcdatetime(value)
            elif key == 'S024':
                header['sh']['ORIGIN'] = to_utcdatetime(value)
            elif key:
                key = INVERTED_SH_IDX.get(key, key)
                if key in SH_KEYS_INT:
                    header['sh'][key] = int(value)
                elif key in SH_KEYS_FLOAT:
                    header['sh'][key] = float(value)
                else:
                    header['sh'][key] = value
        # set channel code
        header['channel'] = ''.join(channel)
        # remember record number
        header['sh']['RECNO'] = len(stream) + 1
        if headonly:
            # skip data
            stream.append(Trace(header=header))
        else:
            if not npts:
                stream.append(Trace(header=header))
                continue
            # read data
            data = fh_data.read(npts * 4)
            dtype = byteorder + 'f4'
            data = from_buffer(data, dtype=dtype)
            # convert to system byte order
            data = np.require(data, '=f4')
            stream.append(Trace(data=data, header=header))
    if not headonly:
        fh_data.close()
    return stream


def _write_q(stream, filename, data_directory=None, byteorder='=',
             append=False, **kwargs):  # @UnusedVariable
    """
    Writes a Seismic Handler Q file from given ObsPy Stream object.

    .. warning::
        This function should NOT be called directly, it registers via the
        the :meth:`~obspy.core.stream.Stream.write` method of an
        ObsPy :class:`~obspy.core.stream.Stream` object, call this instead.

    :type stream: :class:`~obspy.core.stream.Stream`
    :param stream: The ObsPy Stream object to write.
    :type filename: str
    :param filename: Name of the Q file to write.
    :type data_directory: str, optional
    :param data_directory: Data directory where the corresponding QBN will be
        written.
    :type byteorder: str, optional
    :param byteorder: Enforce byte order for data file. Can be little endian
        (``'<'``), big endian (``'>'``), or native byte order (``'='``).
        Defaults to ``'='``.
    :type append: bool, optional
    :param append: If filename exists append all data to file, default False.
    """
    if filename.endswith('.QHD') or filename.endswith('.QBN'):
        path = Path(filename)
        filename = str(Path(path.parent)/path.stem)
    if data_directory:
        filename_data = Path(data_directory) / Path(filename).name
    else:
        filename_data = filename
    filename_header = filename + '.QHD'

    # if the header file exists its assumed that the data is also there
    if Path(filename_header).exists() and append:
        try:
            trcs = _read_q(filename_header, headonly=True)
        except Exception:
            raise Exception("Target filename '%s' not readable!" % filename)
        mode = 'ab'
        count_offset = len(trcs)
        cur_npts_offset = sum([trcs[i].stats.npts for i in range(len(trcs))])
    else:
        append = False
        mode = 'wb'
        count_offset = 0
        cur_npts_offset = 0

    fh = open(filename_header, mode)
    fh_data = open(filename_data + '.QBN', mode)

    # build up header strings
    headers = []
    minnol = 4
    cur_npts = 0 + cur_npts_offset
    for trace in stream:
        temp = "L000:%d~ " % cur_npts
        cur_npts += trace.stats.npts
        temp += "L001:%d~ R000:%f~ R026:%f~ " % (trace.stats.npts,
                                                 trace.stats.delta,
                                                 trace.stats.calib)
        if trace.stats.station:
            temp += "S001:%s~ " % trace.stats.station
        # component must be split
        if len(trace.stats.channel) > 2:
            temp += "C000:%c~ " % trace.stats.channel[2]
        if len(trace.stats.channel) > 0:
            temp += "C001:%c~ " % trace.stats.channel[0]
        if len(trace.stats.channel) > 1:
            temp += "C002:%c~ " % trace.stats.channel[1]
        # special format for start time
        dt = trace.stats.starttime
        temp += "S021:%s~ " % from_utcdatetime(dt)
        for key, value in trace.stats.get('sh', {}).items():
            # skip unknown keys
            if not key or key not in SH_IDX.keys():
                continue
            # convert UTCDateTimes into strings
            if isinstance(value, UTCDateTime):
                value = from_utcdatetime(value)
            temp += "%s:%s~ " % (SH_IDX[key], value)
        headers.append(temp)
        # get maximal number of trclines
        nol = len(temp) // 74 + 1
        if nol > minnol:
            minnol = nol
    # first line: magic number, cmtlines, trclines
    # XXX: comment lines are ignored
    if not append:
        line = "43981 1 %d\n" % minnol
        fh.write(line.encode('ascii', 'strict'))

    for i, trace in enumerate(stream):
        # write headers
        temp = [headers[i][j:j + 74] for j in range(0, len(headers[i]), 74)]
        for j in range(0, minnol):
            try:
                line = "%02d|%s\n" % ((i + 1 + count_offset) % 100, temp[j])
                fh.write(line.encode('ascii', 'strict'))
            except Exception:
                line = "%02d|\n" % ((i + 1 + count_offset) % 100)
                fh.write(line.encode('ascii', 'strict'))
        # write data in given byte order
        dtype = byteorder + 'f4'
        data = np.require(trace.data, dtype=dtype)
        fh_data.write(data.data)
    fh.close()
    fh_data.close()


def to_utcdatetime(value):
    """
    Converts time string used within Seismic Handler into a UTCDateTime.

    :type value: str
    :param value: A Date time string.
    :return: Converted :class:`~obspy.core.utcdatetime.UTCDateTime` object.

    .. rubric:: Example

    >>> to_utcdatetime(' 2-JAN-2008_03:04:05.123')
    UTCDateTime(2008, 1, 2, 3, 4, 5, 123000)
    >>> to_utcdatetime('2-JAN-2008')
    UTCDateTime(2008, 1, 2, 0, 0)
    >>> to_utcdatetime('2-JAN-08')
    UTCDateTime(2008, 1, 2, 0, 0)
    >>> to_utcdatetime('2-JAN-99')
    UTCDateTime(1999, 1, 2, 0, 0)
    >>> to_utcdatetime('2-JAN-2008_1')
    UTCDateTime(2008, 1, 2, 1, 0)
    >>> to_utcdatetime('2-JAN-2008_1:1')
    UTCDateTime(2008, 1, 2, 1, 1)
    """
    try:
        date, time = value.split('_')
    except ValueError:
        date = value
        time = "0:0:0"
    day, month, year = date.split('-')
    time = time.split(':')
    try:
        hour, mins, secs = time
    except ValueError:
        hour = time[0]
        mins = "0"
        secs = "0"
        if len(time) == 2:
            mins = time[1]
    day = int(day)
    try:
        month = MONTHS.index(month.upper()) + 1
    except ValueError:
        month = MONTHS_DE.index(month.upper()) + 1
    if len(year) == 2:
        if int(year) < 70:
            year = "20" + year
        else:
            year = "19" + year
    year = int(year)
    hour = int(hour)
    mins = int(mins)
    secs = float(secs)
    return UTCDateTime(year, month, day, hour, mins) + secs


def from_utcdatetime(dt):
    """
    Converts UTCDateTime object into a time string used within Seismic Handler.

    :type dt: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param dt: A UTCDateTime object.
    :return: Converted date time string usable by Seismic Handler.

    .. rubric:: Example

    >>> from obspy import UTCDateTime
    >>> dt = UTCDateTime(2008, 1, 2, 3, 4, 5, 123456)
    >>> print(from_utcdatetime(dt))
     2-JAN-2008_03:04:05.123
    """
    pattern = "%2d-%3s-%4d_%02d:%02d:%02d.%03d"

    return pattern % (dt.day, MONTHS[dt.month - 1], dt.year, dt.hour,
                      dt.minute, dt.second, dt.microsecond / 1000)


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
The obspy.imaging.spectrogram test suite.
"""
import warnings

import numpy as np
import matplotlib.pyplot as plt
import pytest

from obspy import Stream, Trace, UTCDateTime
from obspy.imaging import spectrogram


class TestSpectrogram:
    """
    Test cases for spectrogram plotting.
    """
    def test_spectrogram(self, image_path):
        """
        Create spectrogram plotting examples in tests/output directory.
        """
        # Create dynamic test_files to avoid dependencies of other modules.
        # set specific seed value such that random numbers are reproducible
        np.random.seed(815)
        head = {
            'network': 'BW', 'station': 'BGLD',
            'starttime': UTCDateTime(2007, 12, 31, 23, 59, 59, 915000),
            'sampling_rate': 200.0, 'channel': 'EHE'}
        tr = Trace(data=np.random.randint(0, 1000, 824), header=head)
        st = Stream([tr])
        # 1 - using log=True
        image_path_1 = image_path.parent / 'spectrogram1.png'
        with warnings.catch_warnings(record=True):
            warnings.resetwarnings()
            with np.errstate(all='warn'):
                spectrogram.spectrogram(
                    st[0].data, log=True, outfile=image_path_1,
                    samp_rate=st[0].stats.sampling_rate, show=False)
        # 2 - using log=False
        image_path_2 = image_path.parent / 'spectrogram2.png'
        spectrogram.spectrogram(st[0].data, log=False, outfile=image_path_2,
                                samp_rate=st[0].stats.sampling_rate,
                                show=False)

    def test_spectrogram_defaults(self):
        """
        Make sure input at varying sampling rates and data lengths does not
        suffer from defaults (wlen etc) being calculated to nonsensical values
        resulting in errors raised from the underlying mlab spectrogram

        wlen (in s) used to be calculated as (samp_rate / 100) which makes no
        sense, it results in very large window sizes for high sampling rates
        and in window lengths that dont even include a two samples for low
        sampling rates like 1 Hz.
        """
        # input of 30 minutes sampled at 1 Hz should definitely be able to have
        # a sensible default for window length, but it used to fail
        data = np.ones(1800)
        spectrogram.spectrogram(data, samp_rate=1, show=False)
        plt.close('all')

    def test_spectrogram_nice_error_messages(self):
        """
        Make sure to have some nice messages on weird input
        """
        # Only 20 samples at 1 Hz and our default tries to window 128 samples,
        # so we dont even have enough data for a single window. This used to
        # fail inside mlab specgram with ugly error message about the input
        data = np.ones(20)
        msg = (r'Input signal too short \(20 samples, window length 6.4 '
               r'seconds, nfft 128 samples, sampling rate 20.0 Hz\)')
        with pytest.raises(ValueError, match=msg):
            spectrogram.spectrogram(data, samp_rate=20.0, show=False)
        # Only 130 samples at 1 Hz and our default tries to window 128 samples,
        # so we dont have enough data for two windows. In principle we could
        # still plot that but our code currently relies on having at least two
        # windows for plotting. It does not seem worthwhile to change the code
        # to do so, so just show a nice error
        data = np.ones(130)
        msg = (r'Input signal too short \(130 samples, window length 6.4 '
               r'seconds, nfft 128 samples, 115 samples window overlap, '
               r'sampling rate 20.0 Hz\)')
        with pytest.raises(ValueError, match=msg):
            spectrogram.spectrogram(data, samp_rate=20.0, show=False)
        plt.close('all')
